"""A module containing tests to demonstrate that tests are working."""


class TestTest(object):
    """Container to demonstrate that tests are running."""

    def test_trivial(self):
        """Demonstrate that tests are running."""
        assert 1 == 1



"""A module containing tests for the library representation of Schemas."""
# pylint: disable=protected-access
from lxml import etree
import pytest
import iati.codelists
import iati.default
import iati.exceptions
import iati.resources
import iati.schemas
import iati.tests.utilities


class TestSchemas(object):
    """A container for tests relating to Schemas."""

    @pytest.fixture(params=[
        {
            "path_func": iati.resources.get_all_activity_schema_paths,
            "schema_class": iati.ActivitySchema
        },
        {
            "path_func": iati.resources.get_all_organisation_schema_paths,
            "schema_class": iati.OrganisationSchema
        }
    ])
    def schema_initialised(self, request, standard_version_optional):
        """Create and return a single ActivitySchema or OrganisationSchema object.

        For use where both ActivitySchema and OrganisationSchema must produce the same result.

        Returns:
            iati.Schema: An activity or organisation Schema that has been initialised.

        """
        schema_path = request.param['path_func'](*standard_version_optional)[0]
        return request.param['schema_class'](schema_path)

    @pytest.mark.parametrize("schema_func, expected_root_element_name", [
        (iati.default.activity_schema, 'iati-activities'),
        (iati.default.organisation_schema, 'iati-organisations')
    ])
    def test_schema_default_attributes(self, standard_version_optional, schema_func, expected_root_element_name):
        """Check a Schema's default attributes are correct."""
        schema = schema_func(*standard_version_optional)

        assert schema.ROOT_ELEMENT_NAME == expected_root_element_name
        assert expected_root_element_name in schema._source_path

    def test_schema_define_from_xsd(self, schema_initialised):
        """Check that a Schema can be generated from an XSD definition."""
        assert isinstance(schema_initialised.codelists, set)
        assert not schema_initialised.codelists
        assert isinstance(schema_initialised.rulesets, set)
        assert not schema_initialised.rulesets

    @pytest.mark.parametrize("schema_func", [
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    @pytest.mark.parametrize('version', iati.constants.STANDARD_VERSIONS)
    def test_schema_get_version(self, schema_func, version):
        """Check that the correct version number is returned by the base classes of iati.schemas.schema._get_version()."""
        schema = schema_func(version)
        result = schema._get_version()

        assert result == version

    def test_schema_unmodified_includes(self, schema_initialised):
        """Check that local elements can be accessed, but imported elements within unmodified Schema includes cannot be accessed.

        lxml does not contain functionality to access elements within imports defined along the lines of:
        `<xsd:include schemaLocation="NAME.xsd" />`

        Todo:
            Simplify asserts.

            Consider consolidating variables shared between multiple tests.

        """
        schema = schema_initialised
        local_element = schema.ROOT_ELEMENT_NAME
        included_element = 'reporting-org'

        include_location_xpath = (iati.constants.NAMESPACE + 'include')
        local_xpath = (iati.constants.NAMESPACE + 'element[@name="' + local_element + '"]')
        included_xpath = (iati.constants.NAMESPACE + 'element[@name="' + included_element + '"]')

        assert schema._schema_base_tree.getroot().find(include_location_xpath).attrib['schemaLocation'] == 'iati-common.xsd'
        assert isinstance(schema._schema_base_tree.getroot().find(local_xpath), etree._Element)
        assert schema._schema_base_tree.getroot().find(included_xpath) is None

    def test_schema_modified_includes(self, schema_initialised):
        """Check that elements within unflattened modified Schema includes cannot be accessed.

        lxml contains functionality to access elements within imports defined along the lines of:
        `<xi:include href="NAME.xsd" parse="xml" />`
        when there is a namespace defined against the root schema element as `xmlns:xi="http://www.w3.org/2001/XInclude"`

        Todo:
            Simplify asserts.

            Consider consolidating variables shared between multiple tests.

        """
        schema = schema_initialised
        local_element = schema.ROOT_ELEMENT_NAME
        included_element = 'reporting-org'

        include_location_xpath = (iati.constants.NAMESPACE + 'include')
        xi_location_xpath = ('{http://www.w3.org/2001/XInclude}include')
        local_xpath = (iati.constants.NAMESPACE + 'element[@name="' + local_element + '"]')
        included_xpath = (iati.constants.NAMESPACE + 'element[@name="' + included_element + '"]')

        include_location = schema._schema_base_tree.getroot().find(include_location_xpath).attrib['schemaLocation']
        tree = schema._change_include_to_xinclude(schema._schema_base_tree)
        xi_node = tree.getroot().find(xi_location_xpath)
        include_node_after = tree.getroot().find(include_location_xpath)

        # check that the new element has been added
        assert isinstance(xi_node, etree._Element)
        assert xi_node.attrib['href'][-len(include_location):] == include_location
        assert xi_node.attrib['parse'] == 'xml'
        assert isinstance(tree.getroot().find(local_xpath), etree._Element)
        assert not isinstance(tree.getroot().find(included_xpath), etree._Element)
        # check that the old element has been removed
        assert include_node_after is None

    def test_schema_flattened_includes(self, schema_initialised):
        """Check that includes are flattened correctly.

        In a full flatten of included elements as `<xi:include href="NAME.xsd" parse="xml" />`, there may be nested `schema` elements and other situations that are not permitted.

        This checks that the flattened xsd is valid and that included elements can be accessed.

        Todo:
            Simplify asserts.

            Consider consolidating variables shared between multiple tests.

            Assert that the flattened XML is a valid Schema.

        """
        schema = schema_initialised
        local_element = schema.ROOT_ELEMENT_NAME
        included_element = 'reporting-org'

        include_location_xpath = (iati.constants.NAMESPACE + 'include')
        xi_location_xpath = ('{http://www.w3.org/2001/XInclude}include')
        local_xpath = (iati.constants.NAMESPACE + 'element[@name="' + local_element + '"]')
        included_xpath = (iati.constants.NAMESPACE + 'element[@name="' + included_element + '"]')

        tree = schema.flatten_includes(schema._schema_base_tree)

        assert tree.getroot().find(include_location_xpath) is None
        assert tree.getroot().find(xi_location_xpath) is None
        assert isinstance(tree.getroot().find(local_xpath), etree._Element)
        assert isinstance(tree.getroot().find(included_xpath), etree._Element)
        assert iati.utilities.convert_tree_to_schema(tree)

    def test_schema_codelists_add(self, schema_initialised):
        """Check that it is possible to add Codelists to the Schema."""
        codelist_name = "a test Codelist name"
        codelist = iati.Codelist(codelist_name)

        schema_initialised.codelists.add(codelist)

        assert len(schema_initialised.codelists) == 1

    def test_schema_codelists_add_twice(self, schema_initialised):
        """Check that it is not possible to add the same Codelist to a Schema multiple times.

        Todo:
            Consider if this test should test against a versioned Codelist.
        """
        codelist_name = "a test Codelist name"
        codelist = iati.Codelist(codelist_name)

        schema_initialised.codelists.add(codelist)
        schema_initialised.codelists.add(codelist)

        assert len(schema_initialised.codelists) == 1

    def test_schema_codelists_add_duplicate(self, schema_initialised):
        """Check that it is not possible to add multiple functionally identical Codelists to a Schema.

        Todo:
            Consider if this test should test against a versioned Codelist.
        """
        codelist_name = "a test Codelist name"
        codelist = iati.Codelist(codelist_name)
        codelist2 = iati.Codelist(codelist_name)

        schema_initialised.codelists.add(codelist)
        schema_initialised.codelists.add(codelist2)

        assert len(schema_initialised.codelists) == 1

    def test_schema_rulesets_add(self, schema_initialised):
        """Check that it is possible to add Rulesets to the Schema.

        Todo:
            Consider if this test should test against a versioned Ruleset.

        """
        ruleset = iati.default.ruleset()

        schema_initialised.rulesets.add(ruleset)

        assert len(schema_initialised.rulesets) == 1

    @pytest.mark.skip(reason='Not implemented')
    def test_schema_rulesets_add_twice(self, schema_initialised):
        """Check that it is not possible to add the same Rulesets to a Schema multiple times.

        Todo:
            Consider if this test should test against a versioned Ruleset.
        """
        raise NotImplementedError

    @pytest.mark.skip(reason='Not implemented')
    def test_schema_rulesets_add_duplicate(self, schema_initialised):
        """Check that it is not possible to add multiple functionally identical Rulesets to a Schema.

        Todo:
            Consider if this test should test against a versioned Ruleset.
        """
        raise NotImplementedError



"""A top-level namespace package for IATI."""
from .codelists import Code, Codelist  # noqa: F401
from .data import Dataset  # noqa: F401
from .rulesets import Rule, Ruleset  # noqa: F401
from .rulesets import RuleAtLeastOne, RuleDateOrder, RuleDependent, RuleNoMoreThanOne, RuleRegexMatches, RuleRegexNoMatches, RuleStartsWith, RuleSum, RuleUnique  # noqa: F401
from .schemas import ActivitySchema, OrganisationSchema  # noqa: F401

__import__('pkg_resources').declare_namespace(__name__)



"""A module containing tests for the library representation of default values."""
import pytest
import iati.codelists
import iati.constants
import iati.default
import iati.schemas
import iati.tests.utilities


class TestDefault(object):
    """A container for tests relating to Default data."""

    @pytest.mark.parametrize("invalid_version", iati.tests.utilities.generate_test_types(['none'], True))
    @pytest.mark.parametrize("func_to_check", [
        iati.default.get_default_version_if_none,
        iati.default.codelists,
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    def test_invalid_version(self, invalid_version, func_to_check):
        """Check that an invalid version causes an error when obtaining default data."""
        with pytest.raises(ValueError):
            func_to_check(invalid_version)


class TestDefaultCodelists(object):
    """A container for tests relating to default Codelists."""

    @pytest.fixture
    def codelist_name(self):
        """Return the name of a valid Codelist."""
        return 'Country'

    @pytest.fixture
    def codelists_with_no_name_codes(self):
        """Return the names of Codelists where Codes do not have names."""
        return ['FileFormat', 'Version']

    @pytest.mark.parametrize("invalid_version", iati.tests.utilities.generate_test_types(['none'], True))
    def test_invalid_version_single_codelist(self, invalid_version, codelist_name):
        """Check that an invalid version causes an error when obtaining a single default Codelist.

        Note:
            This is a separate test since the function takes a parameter other than the `version`.

        """
        with pytest.raises(ValueError):
            iati.default.codelist(codelist_name, invalid_version)

    def test_default_codelist_valid_at_all_versions(self, codelist_name, standard_version_optional):
        """Check that a named default Codelist may be located.

        Todo:
            Check internal values beyond the codelists being the correct type.
        """
        codelist = iati.default.codelist(codelist_name, *standard_version_optional)

        assert isinstance(codelist, iati.Codelist)
        assert codelist.name == codelist_name
        for code in codelist.codes:
            assert isinstance(code, iati.Code)

    @pytest.mark.parametrize("version, codelist_name, expected_type", [
        ('1.04', 'AidTypeFlag', iati.Codelist),
        ('1.05', 'AidTypeFlag', iati.Codelist),
        ('2.01', 'AidTypeFlag', ValueError),
        ('2.02', 'AidTypeFlag', ValueError),
        ('1.04', 'BudgetStatus', ValueError),
        ('1.05', 'BudgetStatus', ValueError),
        ('2.01', 'BudgetStatus', ValueError),
        ('2.02', 'BudgetStatus', iati.Codelist)
    ])
    def test_default_codelist_valid_only_at_some_versions(self, codelist_name, version, expected_type):
        """Check that a codelist that is valid at some version/s is not valid in other versions.

        Example:
            AidTypeFlag was an embedded codelist in v1.04 and v1.05, but is not valid at any version after this.
            For example, BudgetStatus was added as an embedded codelist in v2.02, so is not valid prior to this.
        """
        try:  # Note pytest.raises() is not used here in order to keep this test flexible for parameterization.
            result = iati.default.codelist(codelist_name, version)
        except ValueError as excinfo:
            result = excinfo

        assert isinstance(result, expected_type)

    @pytest.mark.parametrize("name", iati.tests.utilities.generate_test_types(['str'], True))
    def test_default_codelist_invalid_at_all_versions(self, name, standard_version_optional):
        """Check that trying to find a default Codelist with an invalid name raises an error."""
        with pytest.raises(ValueError) as excinfo:
            iati.default.codelist(name, *standard_version_optional)

        assert 'There is no default Codelist in version' in str(excinfo.value)

    def test_default_codelists_type(self, codelist_lengths_by_version):
        """Check that the default Codelists are of the correct type.

        Todo:
            Switch from type-checking to behavior-checking, which is more Pythonic.
        """
        codelists = iati.default.codelists(codelist_lengths_by_version.version)

        assert isinstance(codelists, dict)
        assert len(codelists.values()) == codelist_lengths_by_version.expected_length
        for codelist in codelists.values():
            assert isinstance(codelist, iati.Codelist)
            for code in codelist.codes:
                assert isinstance(code, iati.Code)

    def test_default_codelists_codes_have_name(self, standard_version_optional, codelists_with_no_name_codes):
        """Check that Codelists with Codes that should have names do have names.

        Codes in a Codelist should have a name. This checks that default Codelists have names. A small number of Codelists are excluded because they are known not to have names.

        """
        codelists = iati.default.codelists(*standard_version_optional)
        relevant_codelists = [codelist for codelist in codelists.values() if codelist.name not in codelists_with_no_name_codes]

        for codelist in relevant_codelists:
            for code in codelist.codes:
                assert code.name != ''

    def test_default_codelists_no_name_codes_have_no_name(self, standard_version_optional, codelists_with_no_name_codes):
        """Check that Codelists with Codes that are known to have no name have no name.

        Ideally all Codes would have a name. There are a couple of Codelists where Codes do not. This test is intended to identify the point in time that names are added.

        """
        codelists = iati.default.codelists(*standard_version_optional)
        relevant_codelists = [codelist for codelist in codelists.values() if codelist.name in codelists_with_no_name_codes]

        for codelist in relevant_codelists:
            for code in codelist.codes:
                assert code.name == ''

    def test_codelist_mapping_condition(self):
        """Check that the Codelist mapping file is having conditions read.

        Todo:
            Split into multiple tests.
        """
        mapping = iati.default.codelist_mapping()

        assert mapping['Sector'][0]['condition'] == "@vocabulary = '1' or not(@vocabulary)"
        assert mapping['Version'][0]['condition'] is None

    def test_codelist_mapping_xpath(self, standard_version_optional):
        """Check that the Codelist mapping file is being read for both org and activity mappings.

        Todo:
            Split into multiple tests.
        """
        mapping = iati.default.codelist_mapping(*standard_version_optional)
        currency_xpaths = [currency_mapping['xpath'] for currency_mapping in mapping['Currency']]

        expected_xpaths = [
            '//iati-activity/@default-currency',
            '//iati-activity/budget/value/@currency',
            '//iati-activity/crs-add/loan-status/@currency',
            '//iati-activity/fss/forecast/@currency',
            '//iati-activity/planned-disbursement/value/@currency',
            '//iati-activity/transaction/value/@currency',
            '//iati-organisation/@default-currency',
            '//iati-organisation/total-budget/value/@currency',
            '//iati-organisation/recipient-org-budget/value/@currency',
            '//iati-organisation/recipient-country-budget/value/@currency'
        ]

        for xpath in expected_xpaths:
            assert xpath in currency_xpaths
        assert mapping['InvalidCodelistName'] == []

    def test_default_codelists_length(self, codelist_lengths_by_version):
        """Check that the default Codelists for each version contain the expected number of Codelists."""
        codelists = iati.default.codelists(codelist_lengths_by_version.version)

        assert len(codelists) == codelist_lengths_by_version.expected_length


class TestDefaultRulesets(object):
    """A container for tests relating to default Rulesets."""

    def test_default_ruleset(self, standard_version_optional):
        """Check that the default Ruleset is correct.

        Todo:
            Check internal values beyond the Ruleset being the correct type.

        """
        ruleset = iati.default.ruleset(*standard_version_optional)

        assert isinstance(ruleset, iati.Ruleset)

    def test_default_ruleset_validation_rules_valid(self, schema_ruleset):
        """Check that a fully valid IATI file does not raise any type of error (including rules/rulesets)."""
        data = iati.tests.utilities.load_as_dataset('valid_std_ruleset')
        result = iati.validator.full_validation(data, schema_ruleset)

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_ruleset)
        assert not result.contains_errors()

    @pytest.mark.parametrize("rule_error, invalid_dataset_name, info_text", [
        (
            'err-rule-at-least-one-conformance-fail',
            'ruleset-std/invalid_std_ruleset_missing_sector_element',
            'At least one of `sector` or `transaction/sector` must be present within each `//iati-activity`.'
        ),
        (
            'err-rule-date-order-conformance-fail',
            'ruleset-std/invalid_std_ruleset_bad_date_order',
            '`activity-date[@type=\'1\']/@iso-date` must be chronologically before `activity-date[@type=\'3\']/@iso-date` within each `//iati-activity`.'
        ),
        (
            'err-rule-regex-matches-conformance-fail',
            'ruleset-std/invalid_std_ruleset_bad_identifier',
            'Each instance of `reporting-org/@ref` and `iati-identifier` and `participating-org/@ref` and `transaction/provider-org/@ref` and `transaction/receiver-org/@ref` within each `//iati-activity` must match the regular expression `[^\\/\\&\\|\\?]+`.'  # noqa: disable=E501 # pylint: disable=line-too-long
        ),
        (
            'err-rule-sum-conformance-fail',
            'ruleset-std/invalid_std_ruleset_does_not_sum_100',
            'Within each `//iati-activity`, the sum of values matched at `recipient-country/@percentage` and `recipient-region/@percentage` must be `100`.'
        )
        # Note the Rules relating to 'dependent', 'no_more_than_one', 'regex_no_matches', 'startswith' and 'unique' are not used in the Standard Ruleset.
    ])
    def test_default_ruleset_validation_rules_invalid(self, schema_ruleset, rule_error, invalid_dataset_name, info_text):
        """Check that the expected rule error is detected when validating files containing invalid data for that rule.

        Note:
            The fixed strings being checked here may be a tad annoying to maintain.
            `test_rule_string_output_general` and `test_rule_string_output_specific` in `test_rulesets.py` do something related for Rules. As such, something more generic may work better in the future.

        Todo:
            Consider whether this test should remove all warnings and assert that there is only the expected warning contained within the test file.

            Check that the expected missing elements appear the the help text for the given element.

        """
        data = iati.tests.utilities.load_as_dataset(invalid_dataset_name)
        result = iati.validator.full_validation(data, schema_ruleset)
        errors_for_rule_error = result.get_errors_or_warnings_by_name(rule_error)
        errors_for_ruleset = result.get_errors_or_warnings_by_name('err-ruleset-conformance-fail')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_ruleset)
        assert not iati.validator.is_valid(data, schema_ruleset)
        assert len(errors_for_rule_error) == 1
        assert len(errors_for_ruleset) == 1
        assert info_text in errors_for_rule_error[0].info


class TestDefaultSchemas(object):
    """A container for tests relating to default Schemas."""

    def test_default_activity_schemas(self, standard_version_optional):
        """Check that the default ActivitySchemas are correct.

        Todo:
            Check internal values beyond the schemas being the correct type.
        """
        schema = iati.default.activity_schema(*standard_version_optional)

        assert isinstance(schema, iati.ActivitySchema)

    def test_default_organisation_schemas(self, standard_version_optional):
        """Check that the default ActivitySchemas are correct.

        Todo:
            Check internal values beyond the schemas being the correct type.
        """
        schema = iati.default.organisation_schema(*standard_version_optional)

        assert isinstance(schema, iati.OrganisationSchema)

    @pytest.mark.parametrize("population_status", [[], [True]])
    @pytest.mark.parametrize("schema_func", [
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    def test_default_schemas_populated(self, population_status, schema_func, codelist_lengths_by_version):
        """Check that the default Codelists for each version contain the expected number of Codelists."""
        schema = schema_func(codelist_lengths_by_version.version, *population_status)

        assert len(schema.codelists) == codelist_lengths_by_version.expected_length
        assert len(schema.rulesets) == 1

    @pytest.mark.parametrize("schema_func", [
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    def test_default_schemas_unpopulated(self, schema_func, standard_version_mandatory):
        """Check that the default Codelists for each version contain the expected number of Codelists."""
        schema = schema_func(standard_version_mandatory[0], False)

        assert schema.codelists == set()
        assert schema.rulesets == set()


class TestDefaultModifications(object):
    """A container for tests relating to the ability to modify defaults."""

    @pytest.fixture
    def codelist_name(self):
        """Return the name of a Codelist that exists at all versions of the Standard."""
        return 'Country'

    @pytest.fixture
    def codelist(self, codelist_name):
        """Return a default Codelist that is part of the IATI Standard."""
        return iati.default.codelist(codelist_name)

    @pytest.fixture
    def codelist_non_default(self):
        """Return a Codelist that is not part of the IATI Standard."""
        return iati.Codelist('custom codelist')

    @pytest.fixture
    def new_code(self):
        """Return a Code object that has not been added to a Codelist."""
        return iati.Code('new code value', 'new code name')

    def test_default_codelist_modification(self, codelist_name, new_code, standard_version_optional):
        """Check that a default Codelist cannot be modified by adding Codes to returned lists."""
        default_codelist = iati.default.codelist(codelist_name, *standard_version_optional)
        base_default_codelist_length = len(default_codelist.codes)

        default_codelist.codes.add(new_code)
        unmodified_codelist = iati.default.codelist(codelist_name, *standard_version_optional)

        assert len(default_codelist.codes) == base_default_codelist_length + 1
        assert len(unmodified_codelist.codes) == base_default_codelist_length

    def test_default_codelists_modification(self, codelist_name, new_code, standard_version_optional):
        """Check that default Codelists cannot be modified by adding Codes to returned lists with default parameters."""
        default_codelists = iati.default.codelists(*standard_version_optional)
        codelist_of_interest = default_codelists[codelist_name]
        base_default_codelist_length = len(codelist_of_interest.codes)

        codelist_of_interest.codes.add(new_code)
        unmodified_codelists = iati.default.codelists(*standard_version_optional)
        unmodified_codelist_of_interest = unmodified_codelists[codelist_name]

        assert len(codelist_of_interest.codes) == base_default_codelist_length + 1
        assert len(unmodified_codelist_of_interest.codes) == base_default_codelist_length

    @pytest.mark.parametrize("default_call", [
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    def test_default_x_schema_modification_unpopulated(self, default_call, codelist, standard_version_mandatory):
        """Check that unpopulated default Schemas cannot be modified.

        Note:
            Implementation is by attempting to add a Codelist to the Schema.

        """
        default_schema = default_call(standard_version_mandatory[0], False)
        base_codelist_count = len(default_schema.codelists)

        default_schema.codelists.add(codelist)
        unmodified_schema = default_call(standard_version_mandatory[0], False)

        assert len(default_schema.codelists) == base_codelist_count + 1
        assert len(unmodified_schema.codelists) == base_codelist_count

    @pytest.mark.parametrize("default_call", [
        iati.default.activity_schema,
        iati.default.organisation_schema
    ])
    def test_default_x_schema_modification_populated(self, default_call, codelist_non_default, standard_version_mandatory):
        """Check that populated default Schemas cannot be modified.

        Note:
            Implementation is by attempting to add a Codelist to the Schema.

        """
        default_schema = default_call(standard_version_mandatory[0], True)
        base_codelist_count = len(default_schema.codelists)

        default_schema.codelists.add(codelist_non_default)
        unmodified_schema = default_call(standard_version_mandatory[0], True)

        assert len(default_schema.codelists) == base_codelist_count + 1
        assert len(unmodified_schema.codelists) == base_codelist_count



"""A module containing tests for the library representation of IATI data.

Todo:
    Implement tests for strict checking once validation work is underway.
"""
import collections
import math
from future.standard_library import install_aliases
from lxml import etree
import pytest
import iati.data
import iati.default
import iati.tests.utilities

install_aliases()


class TestDatasets(object):
    """A container for tests relating to Datasets."""

    @pytest.fixture
    def dataset_initialised(self):
        """Return an initialised Dataset to work from in other tests."""
        return iati.tests.utilities.load_as_dataset('valid_not_iati')

    def test_dataset_no_params(self):
        """Test Dataset creation with no parameters."""
        with pytest.raises(TypeError) as excinfo:
            iati.Dataset()  # pylint: disable=E1120

        assert ('__init__() missing 1 required positional argument' in str(excinfo.value)) or ('__init__() takes exactly 2 arguments' in str(excinfo.value))

    def test_dataset_empty_string(self):
        """Test Dataset creation with an empty string."""
        with pytest.raises(ValueError):
            _ = iati.Dataset('')

    def test_dataset_valid_xml_string(self):
        """Test Dataset creation with a valid XML string that is not IATI data."""
        xml_str = iati.tests.utilities.load_as_string('valid_not_iati')
        data = iati.Dataset(xml_str)

        assert data.xml_str == xml_str.strip()
        assert etree.tostring(data.xml_tree) == etree.tostring(iati.tests.utilities.XML_TREE_VALID)

    def test_dataset_xml_string_leading_whitespace(self):
        """Test Dataset creation with a valid XML string that is not IATI data."""
        xml_str = iati.tests.utilities.load_as_string('leading_whitespace_xml')
        data = iati.Dataset(xml_str)
        tree = etree.fromstring(xml_str.strip())

        assert data.xml_str == xml_str.strip()
        assert etree.tostring(data.xml_tree) == etree.tostring(tree)

    def test_dataset_valid_iati_string(self):
        """Test Dataset creation with a valid IATI XML string."""
        pass

    def test_dataset_invalid_xml_string(self):
        """Test Dataset creation with a string that is not valid XML."""
        with pytest.raises(iati.exceptions.ValidationError) as excinfo:
            iati.Dataset(iati.tests.utilities.load_as_string('invalid'))

        assert excinfo.value.error_log.contains_error_called('err-not-xml-empty-document')

    @pytest.mark.parametrize("not_xml", iati.tests.utilities.generate_test_types(['str'], True))
    def test_dataset_number_not_xml(self, not_xml):
        """Test Dataset creation when it's passed a number rather than a string or etree."""
        with pytest.raises(TypeError) as excinfo:
            iati.Dataset(not_xml)

        assert 'Datasets can only be ElementTrees or strings containing valid XML, using the xml_tree and xml_str attributes respectively. Actual type:' in str(excinfo.value)

    def test_dataset_tree(self):
        """Test Dataset creation with an etree that is not valid IATI data."""
        tree = iati.tests.utilities.XML_TREE_VALID
        data = iati.Dataset(tree)

        assert etree.tostring(data.xml_tree, pretty_print=True) == etree.tostring(tree, pretty_print=True)
        assert data.xml_str == etree.tostring(tree, pretty_print=True)

    def test_dataset_iati_tree(self):
        """Test Dataset creation with a valid IATI etree.

        Todo:
            Implement this function.

        """
        pass

    def test_dataset_xml_str_assignment_valid_str(self, dataset_initialised):
        """Test assignment to the xml_str property with a valid XML string.

        Todo:
            Check that the tree is updated correctly.

        """
        xml_str = iati.tests.utilities.load_as_string('valid_not_iati')
        data = dataset_initialised
        data.xml_str = xml_str

        assert data.xml_str == xml_str.strip()

    def test_dataset_xml_str_assignment_invalid_str(self, dataset_initialised):
        """Test assignment to the xml_str property with an invalid XML string."""
        xml_str = iati.tests.utilities.load_as_string('invalid')
        data = dataset_initialised

        with pytest.raises(iati.exceptions.ValidationError) as excinfo:
            data.xml_str = xml_str

        excinfo.value.error_log.contains_error_called('err-not-xml-empty-document')

    def test_dataset_xml_str_assignment_tree(self, dataset_initialised):
        """Test assignment to the xml_str property with an ElementTree."""
        data = dataset_initialised

        with pytest.raises(TypeError) as excinfo:
            data.xml_str = iati.tests.utilities.XML_TREE_VALID

        assert str(excinfo.value) == 'If setting a Dataset with an ElementTree, use the xml_tree property, not the xml_str property.'

    @pytest.mark.parametrize("invalid_value", iati.tests.utilities.generate_test_types(['str'], True))
    def test_dataset_xml_str_assignment_invalid_value(self, dataset_initialised, invalid_value):
        """Test assignment to the xml_str property with a value that is very much not valid."""
        data = dataset_initialised

        with pytest.raises(TypeError) as excinfo:
            data.xml_str = invalid_value

        assert 'Datasets can only be ElementTrees or strings containing valid XML, using the xml_tree and xml_str attributes respectively. Actual type:' in str(excinfo.value)

    def test_dataset_xml_tree_assignment_valid_tree(self, dataset_initialised):
        """Test assignment to the xml_tree property with a valid ElementTree.

        Todo:
            Check that the xml_tree attribute is updated to the new tree.

        """
        data = dataset_initialised
        data.xml_tree = iati.tests.utilities.XML_TREE_VALID

        assert data.xml_str == etree.tostring(iati.tests.utilities.XML_TREE_VALID, pretty_print=True)

    def test_dataset_xml_tree_assignment_invalid_tree(self, dataset_initialised):
        """Test assignment to the xml_tree property with an invalid ElementTree.

        Todo:
            Create an invalid tree and test it.

        """
        pass

    def test_dataset_xml_tree_assignment_str(self, dataset_initialised):
        """Test assignment to the xml_tree property with an XML string."""
        xml_str = iati.tests.utilities.load_as_string('valid_not_iati')
        data = dataset_initialised

        with pytest.raises(TypeError) as excinfo:
            data.xml_tree = xml_str

        assert 'If setting a Dataset with the xml_property, an ElementTree should be provided, not a' in str(excinfo.value)

    @pytest.mark.parametrize("invalid_value", iati.tests.utilities.generate_test_types(['str'], True))
    def test_dataset_xml_tree_assignment_invalid_value(self, dataset_initialised, invalid_value):
        """Test assignment to the xml_tree property with a value that is very much not valid."""
        data = dataset_initialised

        with pytest.raises(TypeError) as excinfo:
            data.xml_tree = invalid_value

        assert 'If setting a Dataset with the xml_property, an ElementTree should be provided, not a' in str(excinfo.value)


class TestDatasetWithEncoding(object):
    """A container for tests relating to creating a Dataset from various types of input.

    This may be files vs strings, or may revolve around character encoding.

    """

    @pytest.fixture
    def xml_needing_encoding(self):
        """An XML string with a placeholder for an encoding through use of `str.format()`"""
        xml = """<?xml version="1.0" encoding="{}"?>
        <iati-activities version="xx">
          <iati-activity>
             <iati-identifier></iati-identifier>
         </iati-activity>
        </iati-activities>"""

        return xml

    def test_instantiation_dataset_from_string(self):
        """Test that a Dataset instantiated directly from a string (rather than a file) correctly creates an iati.data.Dataset and the input data is contained within the object."""
        xml_str = """<?xml version="1.0"?>
        <iati-activities version="xx">
          <iati-activity>
             <iati-identifier></iati-identifier>
         </iati-activity>
        </iati-activities>"""

        dataset = iati.data.Dataset(xml_str)

        assert isinstance(dataset, iati.data.Dataset)
        assert dataset.xml_str == xml_str

    @pytest.mark.parametrize("encoding", [
        "UTF-8",
        "utf-8",
        "UTF-16",
        "utf-16",
        "UTF-32",
        "utf-32",
        "ASCII",
        "ISO-8859-1",
        "ISO-8859-2",
        "BIG5",
        "EUC-JP"
    ])
    def test_instantiation_dataset_from_string_with_encoding(self, xml_needing_encoding, encoding):
        """Test that an encoded Dataset instantiated directly from a string (rather than a file) correctly creates an iati.data.Dataset and the input data is contained within the object.

        Note:
            The use of UTF-8 and UTF-16 is strongly recommended for IATI datasets, however other encodings are specificed here to demonstrate compatibility.

        """
        xml = xml_needing_encoding.format(encoding)
        xml_encoded = xml.encode(encoding)  # Encode the whole string in line with the specified encoding

        dataset = iati.data.Dataset(xml_encoded)

        assert isinstance(dataset, iati.data.Dataset)
        assert dataset.xml_str == xml_encoded

    @pytest.mark.parametrize("encoding_declared, encoding_used", [
        ("UTF-16", "UTF-8"),
        ("UTF-16", "ISO-8859-1"),
        ("UTF-16", "ASCII"),
        ("UTF-16", "BIG5"),
        ("UTF-16", "EUC-JP")
    ])
    def test_instantiation_dataset_from_string_with_encoding_mismatch(self, xml_needing_encoding, encoding_declared, encoding_used):
        """Test that an error is raised when attempting to create a Dataset where a string is encoded significantly differently from what is defined within the XML encoding declaration.

        Todo:
            Amend error message, when the todo in iati.data.Dataset.xml_str() has been resolved.

        Note:
            There are a number of other errors that may be raised with alternative encoding mismatches. These are not supported since it does not appear likely enough that they will occur and be a large issue in practice.

            This is due to a pair of issues with libxml2 (the underlying library behind lxml):

            1. It only supports a limited number of encodings out-of-the-box.
            2. Different encoding pairs (whether supported or unsupported by libxml2; byte-equivalent-subsets or distinct encodings; and more), will return different error codes in what one would expect to act as equivalent situations.

        """
        xml = xml_needing_encoding.format(encoding_declared)
        xml_encoded = xml.encode(encoding_used)  # Encode the whole string in line with the specified encoding

        with pytest.raises(iati.exceptions.ValidationError) as excinfo:
            _ = iati.data.Dataset(xml_encoded)

        assert excinfo.value.error_log.contains_error_called('err-encoding-invalid')

    @pytest.mark.parametrize("encoding", ["CP424"])
    def test_instantiation_dataset_from_string_with_unsupported_encoding(self, xml_needing_encoding, encoding):
        """Test that an error is raised when attempting to create a dataset where a string is encoded significantly differently from what is defined within the XML encoding declaration.

        Todo:
            Amend error message, when the todo in iati.data.Dataset.xml_str() has been resolved.

        """
        xml = xml_needing_encoding.format(encoding)
        xml_encoded = xml.encode(encoding)  # Encode the whole string in line with the specified encoding

        with pytest.raises(iati.exceptions.ValidationError) as excinfo:
            _ = iati.data.Dataset(xml_encoded)

        assert excinfo.value.error_log.contains_error_called('err-encoding-unsupported')


class TestDatasetSourceFinding(object):
    """A container for tests relating to finding source context within a Dataset."""

    @pytest.fixture(params=[
        iati.tests.utilities.load_as_dataset('valid_not_iati'),
        iati.tests.utilities.load_as_dataset('valid_iati')
    ])
    def data(self, request):
        """A Dataset to test."""
        return request.param

    @pytest.fixture
    def split_xml_str(self, data):
        """The XML from the provided Dataset, split by line."""
        return [''] + data.xml_str.split('\n')

    @pytest.fixture
    def num_lines_xml(self, split_xml_str):
        """The number of lines in the XML string."""
        return len(split_xml_str)

    def test_dataset_xml_str_source_at_line_valid_line_number(self, data, split_xml_str):
        """Test obtaining source of a particular line. Line numbers are valid."""
        for idx, line in enumerate(split_xml_str):
            assert data.source_at_line(idx) == line.strip()

    @pytest.mark.parametrize("line_el_pair", [
        {'line': 3, 'el': '//parent'},
        {'line': 4, 'el': '//child'},
        {'line': 5, 'el': '//another-child'},
        {'line': 7, 'el': '//sub-child'}
    ])
    def test_dataset_xml_str_source_at_line_matches_tree(self, line_el_pair):
        """Test obtaining source of a particular line. Line numbers are valid.

        Ensure that the line numbers from which source is being returned are the same ones provided by the `sourceline` attribute from tree elements.
        """
        data = iati.tests.utilities.load_as_dataset('valid_not_iati')
        split_xml_str = [''] + data.xml_str.split('\n')
        line_num = line_el_pair['line']
        el_from_tree = data.xml_tree.xpath(line_el_pair['el'])[0]
        str_from_tree = etree.tostring(el_from_tree, pretty_print=True).strip().decode('utf-8').split('\n')[0]

        assert el_from_tree.sourceline == line_num
        assert data.source_at_line(line_num) == str_from_tree
        assert data.source_at_line(line_num) == split_xml_str[line_num].strip()

    def test_dataset_xml_str_source_at_line_invalid_line_number(self, data, num_lines_xml):
        """Test obtaining source of a particular line. Line numbers are not valid."""
        with pytest.raises(ValueError):
            data.source_at_line(-1)

        with pytest.raises(ValueError):
            data.source_at_line(num_lines_xml)

    @pytest.mark.parametrize("invalid_value", iati.tests.utilities.generate_test_types(['int'], True))
    def test_dataset_xml_str_source_at_line_invalid_line_type(self, invalid_value, data):
        """Test obtaining source of a particular line. Line numbers are not valid."""
        with pytest.raises(TypeError):
            data.source_at_line(invalid_value)

    def test_dataset_xml_str_source_around_line_valid_line_number(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The line is in the middle of an XML document so that there will be full context both before and after the specified line number.
        Line numbers are valid.
        Uses the default number of surrounding context lines.
        """
        for line_num in range(2, num_lines_xml):
            desired_source = '\n'.join(split_xml_str[line_num - 1:line_num + 2])
            actual_source = data.source_around_line(line_num)

            assert actual_source == desired_source

    def test_dataset_xml_str_source_around_line_valid_line_number_custom_context(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The lines are in the middle of an XML document so that there will be full context both before and after the specified line number.
        Line numbers are valid.
        Uses a custom number of surrounding context lines.
        """
        for context_lines in range(1, math.ceil(num_lines_xml / 2)):
            for line_num in range(context_lines, num_lines_xml - context_lines):
                desired_source = '\n'.join(split_xml_str[max(line_num - context_lines, 1):line_num + context_lines + 1])
                actual_source = data.source_around_line(line_num, context_lines)

                assert actual_source == desired_source

    def test_dataset_xml_str_source_around_line_first_line(self, data, split_xml_str):
        """Test obtaining source around a particular line.

        The line is at the start of an XML document such that there will not be full context before the specified line, but will be afterwards.
        Line numbers are valid.
        Uses the default number of surrounding context lines.
        """
        assert data.source_around_line(0) == '\n'.join(split_xml_str[1:2])

    def test_dataset_xml_str_source_around_line_early_line_custom_context(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The lines are around the start of an XML document such that there will not be full context before the specified line, but will be afterwards.
        Line numbers are valid.
        Uses a custom number of surrounding context lines.
        """
        for context_lines in range(1, math.ceil(num_lines_xml / 2)):
            for line_num in range(0, context_lines):
                desired_source = '\n'.join(split_xml_str[1:line_num + context_lines + 1])
                actual_source = data.source_around_line(line_num, context_lines)

                assert actual_source == desired_source

    def test_dataset_xml_str_source_around_line_last_line(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The line is at the end of an XML document such that there will not be full context after the specified line, but will be before.
        Line numbers are valid.
        Uses the default number of surrounding context lines.
        """
        assert data.source_around_line(num_lines_xml - 1) == '\n'.join(split_xml_str[-2:])

    def test_dataset_xml_str_source_around_line_late_line_custom_context(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The lines are around the end of an XML document such that there will not be full context after the specified line, but will be before.
        Line numbers are valid.
        Uses the default number of surrounding context lines.
        """
        for context_lines in range(1, math.ceil(num_lines_xml / 2)):
            for line_num in range(0, context_lines):
                desired_source = '\n'.join(split_xml_str[-(line_num + context_lines + 1):])
                actual_source = data.source_around_line(num_lines_xml - line_num - 1, context_lines)

                assert actual_source == desired_source

    def test_dataset_xml_str_source_around_line_single_line(self, data, split_xml_str, num_lines_xml):
        """Test obtaining source around a particular line.

        The context is such that only the specified line will be returned.
        """
        for line_num in range(0, num_lines_xml):
            assert data.source_around_line(line_num, 0) == split_xml_str[line_num]
            assert data.source_around_line(line_num, 0).strip() == data.source_at_line(line_num)

    def test_dataset_xml_str_source_around_line_full_file(self, data, num_lines_xml):
        """Test obtaining source around a particular line.

        The context is such that the full file will be returned.
        """
        line_num = int(num_lines_xml / 2)
        context_lines = num_lines_xml

        assert data.source_around_line(line_num, context_lines) == data.xml_str

    def test_dataset_xml_str_source_around_line_negative_context_lines(self, data, num_lines_xml):
        """Test obtaining source around a particular line.

        The number of context lines is negative.
        """
        for line_num in range(0, num_lines_xml):
            with pytest.raises(ValueError):
                data.source_around_line(line_num, -1)

    @pytest.mark.parametrize("invalid_value", iati.tests.utilities.generate_test_types(['int'], True))
    def test_dataset_xml_str_source_around_line_invalid_context_lines(self, invalid_value, data, num_lines_xml):
        """Test obtaining source of a particular line.

        The specified number of context lines is not an integer.
        """
        for line_num in range(0, num_lines_xml):
            with pytest.raises(TypeError):
                data.source_around_line(line_num, invalid_value)


class TestDatasetVersionDetection(object):
    """A container for tests relating to detecting the version of a Dataset."""

    @pytest.fixture(params=[
        ('iati-activities', 'iati-activity'),
        ('iati-organisations', 'iati-organisation')
    ])
    def iati_tag_names(self, request):
        """Return the tag names for an activity or organisaion dataset."""
        output = collections.namedtuple('output', 'root_element child_element')
        return output(root_element=request.param[0], child_element=request.param[1])

    @pytest.mark.parametrize("version", iati.utilities.versions_for_integer(1))
    def test_detect_version_v1_simple(self, iati_tag_names, version):
        """Check that a version 1 Dataset is detected correctly.
        Also checks that version numbers containing whitespace do not affect version detection.
        """
        data = iati.Dataset("""
        <{0} version="{2}">
            <{1} version="{2}"></{1}>
            <{1} version="{2}  "></{1}>
            <{1} version="   {2}"></{1}>
            <{1} version="   {2}   "></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element, version))
        result = data.version

        assert result == version

    def test_detect_version_explicit_parent_mismatch_explicit_child(self, iati_tag_names):
        """Check that no version is detected for a v1 Dataset where a version within the `iati-activities` element does not match the versions specified within all `iati-activity` child elements."""
        data = iati.Dataset("""
        <{0} version="1.02">
            <{1} version="1.02"></{1}>
            <{1} version="1.03"></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result is None

    def test_detect_version_implicit_parent_matches_implicit_child(self, iati_tag_names):
        """Check that the default version is detected for a Dataset where no versions are declared (i.e. the default version is assumed for all `iati-activities` and `iati-activity` child elements)."""
        data = iati.Dataset("""
        <{0}>
            <{1}></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result == '1.01'

    def test_detect_version_explicit_parent_matches_implicit_child(self, iati_tag_names):
        """Check that the default version is detected for a Dataset with the default version explicitly defined at `iati-activities` level, but where all `iati-activity` child elements are not defined (i.e. the default version is assumed)."""
        data = iati.Dataset("""
        <{0} version="1.01">
            <{1}></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result == '1.01'

    def test_detect_version_implicit_parent_matches_explicit_and_implicit_child(self, iati_tag_names):
        """Check that the default version is detected for a Dataset with no version not defined at `iati-activities` level (i.e. the default version is assumed), but where at least one `iati-activity` child element has the default version defined."""
        data = iati.Dataset("""
        <{0}>
            <{1} version="1.01"></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result == '1.01'

    def test_detect_version_explicit_parent_mismatch_implicit_child(self, iati_tag_names):
        """Check that no version is detected for a Dataset that has a non-default version defined at the `iati-activities` level, but no version is defined in any `iati-activity` child element (i.e. the default version is assumed)."""
        data = iati.Dataset("""
        <{0} version="1.02">
            <{1}></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result is None

    def test_detect_version_imlicit_parent_mismatch_explicit_child(self, iati_tag_names):
        """Check that no version is detected for a Dataset that has no version defined at the `iati-activities` level (i.e. the default version is assumed), but at least one non-default version is defined in any `iati-activity` child element."""
        data = iati.Dataset("""
        <{0}>
            <{1} version="1.02"></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element))
        result = data.version

        assert result is None

    @pytest.mark.parametrize("version", iati.utilities.versions_for_integer(2))
    def test_detect_version_v2_simple(self, iati_tag_names, version):
        """Check that a version 2 Dataset is detected correctly."""
        data = iati.Dataset("""
        <{0} version="{2}">
            <{1}></{1}>
            <{1}></{1}>
        </{0}>
        """.format(iati_tag_names.root_element, iati_tag_names.child_element, version))
        result = data.version

        assert result == version

    def test_cannot_assign_to_version_property(self):
        """Check that it is not possible to assign to the `version` property."""
        data = iati.resources.load_as_dataset(iati.resources.get_test_data_path('valid_iati'))

        with pytest.raises(AttributeError) as excinfo:
            data.version = 'test'

        assert "can't set attribute" in str(excinfo.value)



"""A module containing a core representation of IATI Schemas."""
from lxml import etree
import iati.codelists
import iati.constants
import iati.exceptions
import iati.resources
import iati.utilities


class Schema(object):
    """Representation of a Schema as defined within the IATI SSOT. This is used as a base class for ActivitySchema and OrganisationSchema and should not be instantiated directly.

    Attributes:
        codelists (set): The Codelists associated with this Schema.
        rulesets (set): The Rulesets associated with this Schema.
        ROOT_ELEMENT_NAME (str): The name of the root element within the XML Schema that the class represents.

    Warning:
        The private attribute allowing access to the base Schema Tree is likely to change in determining a good way of accessing the contained schema content.

    Todo:
        Determine a good API for accessing the XMLSchema that the iati.Schema represents.

    """

    ROOT_ELEMENT_NAME = ''

    def __init__(self, path):
        """Initialise a Schema.

        Args:
            path (str): The path to the Schema that is being initialised.

        Raises:
            iati.exceptions.SchemaError: An error occurred during the creation of the Schema.

        Warning:
            The format of the constructor is likely to change. It needs to be less reliant on the name acting as a UID, and allow for other attributes to be provided at this point.

            The raised exceptions are likely to change upon review of IATI-specific exceptions.

            Need to define a good API for accessing public and private attributes. Requiring something along the lines of `schema.schema` is likely not ideal. An improved understanding of use cases will be required for this.

        Todo:
            Better use the try-except pattern.

            Allow the base schema to be modified after initialisation.

            Create test instance where the SchemaError is raised.

        """
        self._schema_base_tree = None
        self._source_path = path
        self.codelists = set()
        self.rulesets = set()

        try:
            loaded_tree = iati.resources.load_as_tree(path)
        except OSError:
            msg = "Failed to load tree at '{0}' when creating Schema.".format(path)
            iati.utilities.log_error(msg)
            raise iati.exceptions.SchemaError
        else:
            self._schema_base_tree = loaded_tree

    def _change_include_to_xinclude(self, tree):
        """Change the method in which common elements are included.

        lxml does not contain functionality to access elements within imports defined along the lines of: `<xsd:include schemaLocation="NAME.xsd" />`
        It does, however, contains functionality to access elements within imports defined along the lines of: `<xi:include href="NAME.xsd" parse="xml" />`
        when there is a namespace defined against the root schema element as `xmlns:xi="http://www.w3.org/2001/XInclude"`

        This changes instances of the former to the latter.

        Args:
            tree (etree._ElementTree): The tree within which xsd:include is to be changed to xi:include.

        Returns:
            etree._ElementTree: The modified tree.

        Todo:
            Add more robust tests for schemas at different versions.

            Check whether this is safe in the general case, so allowing it to be performed in __init__().

            Make resource locations more able to handle the general case.

            Consider moving this out of Schema().

            Tidy this up.

            Consider using XSLT.

        """
        # identify the old info
        include_xpath = (iati.constants.NAMESPACE + 'include')
        include_el = tree.getroot().find(include_xpath)
        if include_el is None:
            return
        include_location = include_el.attrib['schemaLocation']

        # add namespace for XInclude
        xi_name = 'xi'
        xi_uri = 'http://www.w3.org/2001/XInclude'
        iati.utilities.add_namespace(tree, xi_name, xi_uri)
        new_nsmap = {}
        for key, value in iati.constants.NSMAP.items():
            new_nsmap[key] = value
        new_nsmap[xi_name] = xi_uri

        # create a new element
        xinclude_el = etree.Element(
            '{' + xi_uri + '}include',
            href=iati.resources.resource_filename(iati.resources.get_schema_path(include_location[:-4], self._get_version())),
            parse='xml',
            nsmap=new_nsmap
        )

        # make the path to `xml.xsd` reference the correct file
        import_xpath = (iati.constants.NAMESPACE + 'import')
        import_el = tree.getroot().find(import_xpath)
        import_el.attrib['schemaLocation'] = iati.resources.resource_filename(iati.resources.get_schema_path('xml', self._get_version()))

        # insert the new element
        tree.getroot().insert(import_el.getparent().index(import_el) + 1, xinclude_el)

        # remove the old element
        etree.strip_elements(tree.getroot(), include_xpath)

        return tree

    def _get_version(self):
        """Return the version that this schema is defined as.

        Returns:
            str or None: The version stated for the schema, according to the value defined in the 'version' attribute at root of the XSD schema. Returns None if there is no 'version' attribute.

        """
        return self._schema_base_tree.getroot().get('version')

    def flatten_includes(self, tree):
        """Flatten includes so that all nodes are accessible through lxml.

        Identify the contents of files defined as `<xsd:include schemaLocation="NAME.xsd" />` and bring in the contents.

        Args:
            tree (etree._ElementTree): The tree to flatten.

        Returns:
            etree._ElementTree: The flattened tree.

        Todo:
            Add more robust tests for schemas at different versions.

            Consider moving this out of Schema().

            Tidy this up.

        """
        # change the include to a format that lxml can read
        tree = self._change_include_to_xinclude(tree)

        # adopt the included elements
        tree.xinclude()

        # remove nested schema elements
        schema_xpath = (iati.constants.NAMESPACE + 'schema')
        for nested_schema_el in tree.getroot().findall(schema_xpath):
            if isinstance(nested_schema_el, etree._Element):  # pylint: disable=protected-access
                # move contents of nested schema elements up a level
                for elem in nested_schema_el[:]:
                    # do not duplicate an import statement
                    if 'schemaLocation' in elem.attrib:
                        continue
                    tree.getroot().insert(nested_schema_el.getparent().index(nested_schema_el) + 1, elem)
        # remove the nested schema elements
        etree.strip_elements(tree.getroot(), schema_xpath)

        return tree

    def validator(self):
        """Return a schema that can be used for validation.

        Takes the base schema and converts it into an object that lxml can deal with.

        Returns:
            etree.XMLSchema: A schema that can be used for validation.

        Raises:
            iati.exceptions.SchemaError: An error occurred in the creation of the validator.

        """
        try:
            return iati.utilities.convert_tree_to_schema(self._schema_base_tree)
        except etree.XMLSchemaParseError as err:
            iati.utilities.log_error(err)
            raise iati.exceptions.SchemaError('Problem parsing Schema')


class ActivitySchema(Schema):
    """Representation of an IATI Activity Schema as defined within the IATI SSOT."""

    ROOT_ELEMENT_NAME = 'iati-activities'


class OrganisationSchema(Schema):
    """Representation of an IATI Organisation Schema as defined within the IATI SSOT."""

    ROOT_ELEMENT_NAME = 'iati-organisations'



"""A module containing constants required throughout IATI library code.

The contents of this file are not designed to be user-editable. Only edit if you know what you are doing!

Warning:
    This contents of this module should currently be deemed private.

Todo:
    Allow logging constants to be user-definable.

"""

STANDARD_VERSIONS = ['1.04', '1.05', '2.01', '2.02']
"""Define all versions of the Standard.

Todo:
    This constant to be populated by the values in the Version codelist, rather than hard-coded.

    Consider if functionality should extend to working with development versions of the Standard (e.g. during an upgrade process).

"""

STANDARD_VERSION_LATEST = max(STANDARD_VERSIONS)
"""The latest version of the IATI Standard."""

STANDARD_VERSIONS_MAJOR = list(set([
    int(version.split('.', 1)[0]) for version in STANDARD_VERSIONS
]))
"""The major versions of the IATI Standard."""

LOG_FILE_NAME = 'iatilib.log'
"""The location of the primary IATI log file.

Warning:
    Logging should be clearly user-definable.

"""
LOGGER_NAME = 'iati'
"""The name of the primary IATI Logger.

Warning:
    This should be better based on specific packages.

"""

NAMESPACE = '{http://www.w3.org/2001/XMLSchema}'
"""The namespace that IATI Schema XSD files are specified within."""
NSMAP = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
"""A dictionary for interpreting namespaces in IATI Schemas."""



"""A module containing validation functionality."""

import sys
from lxml import etree
import yaml
import iati.default
import iati.resources


class ValidationError(object):
    """A base class to encapsulate information about Validation Errors."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, err_name, calling_locals=None):
        """Create a new ValidationError.

        Args:
            err_name (str): The name of the error to use as a base.
            calling_locals (dict): The dictionary of local variables from the calling scope. Obtained by calling `locals()`. Default is an empty dictionary.

        Raises:
            ValueError: If there is no base error with the provided name.

        Todo:
            Split message formatting into a child class and raise an error when variables are missing.

            Determine what defaults for attributes should be when the appropriate values are not available.

        """
        # have to set here to ensure each ValidationError has its own dictionary
        if calling_locals is None:
            calling_locals = dict()

        try:
            err_detail = get_error_codes()[err_name]
        except (KeyError, TypeError):
            raise ValueError('{err_name} is not a known type of ValidationError.'.format(**locals()))

        # set general attributes for this type of error
        self.name = err_name
        self.actual_value = None

        for key, val in err_detail.items():
            setattr(self, key, val)

        self.status = 'error' if err_name.split('-')[0] == 'err' else 'warning'

        # format error messages with context-specific info
        try:
            self.help = self.help.format(**calling_locals)
            self.info = self.info.format(**calling_locals)
        except KeyError:  # as missing_var_err:
            # raise NameError('The calling scope must contain a `{missing_var_err.args[0]}` variable for providing information for the error message.'.format(**locals()))
            pass

        # set general attributes for this type of error that require context from the calling scope
        try:
            self.line_number = calling_locals['line_number']
            self.context = calling_locals['dataset'].source_around_line(self.line_number)
        except KeyError:
            pass
        try:
            self.column_number = calling_locals['column_number']
        except KeyError:
            pass
        try:
            self.err = calling_locals['err']
            self.lxml_err_code = calling_locals['err'].type_name
        except (AttributeError, KeyError):
            pass


class ValidationErrorLog(object):
    """A container to keep track of a set of ValidationErrors.

    This acts as an iterable that ValidationErrors can be looped over.

    ValidationErrors may be added to the log.

    Warning:
        It is highly likely that the methods available on a `ValidationErrorLog` will change name. At present the mix of errors, warnings and the combination of the two is confusing. This needs rectifying.

    Todo:
        Make the mix of errors, warnings and both returned by functions clearer, while not being hugely long-winded (`errors_and_warnings`-esque).

    """

    def __init__(self):
        """Initialise the error log."""
        self._values = []

    def __iter__(self):
        """Return an iterator."""
        return iter(self._values)

    def __len__(self):
        """Return the number of items in the ErrorLog."""
        return len(self._values)

    def __getitem__(self, key):
        """Return an item with the specified key."""
        return self._values[key]

    def __eq__(self, other):
        """Test equality with another object."""
        if len(self._values) != len(other):
            return False

        for val in self._values:
            if val not in other:
                return False

        return True

    def add(self, value):
        """Add a single ValidationError to the Error Log.

        Args:
            value (iati.validator.ValidationError): The ValidationError to add to the Error Log.

        Raises:
            TypeError: When attempting to set an item that is not a ValidationError.

        """
        if not isinstance(value, iati.validator.ValidationError):
            raise TypeError('Only ValidationErrors may be added to a ValidationErrorLog.')

        self._values.append(value)

    def contains_error_called(self, err_name):
        """Check the log for an error or warning with the specified name.

        Args:
            err_name (str): The name of the error to look for.

        Returns:
            bool: Whether there is an error or warning with the specified name within the log.

        """
        errors_with_name = self.get_errors_or_warnings_by_name(err_name)

        return len(errors_with_name) > 0

    def contains_error_of_type(self, err_type):
        """Check the log for an error or warning with the specified base exception type.

        Args:
            err_type (type): The type of the error to look for.

        Returns:
            bool: Whether there is an error or warning with the specified type within the log.

        """
        errors_with_type = self.get_errors_or_warning_by_type(err_type)

        return len(errors_with_type) > 0

    def contains_errors(self):
        """Determine whether there are errors contained within the ErrorLog.

        Note:
            The error log may contain warnings, or may be empty.

        Returns:
            bool: Whether there are errors within this error log.

        """
        errors = self.get_errors()

        return len(errors) > 0

    def contains_warnings(self):
        """Determine whether there are warnings contained within the ErrorLog.

        Note:
            The error log may contain errors, or may be empty.

        Returns:
            bool: Whether there are warnings within this error log.

        """
        warnings = self.get_warnings()

        return len(warnings) > 0

    def extend(self, values):
        """Extend the ErrorLog with ValidationErrors from an iterable.

        Args:
            values (iterable): An iterable containing ValidationErrors.

        Note:
            All ValidationErrors within the iterable shall be added. Any other contents shall not, and will fail to be added silently.

        Raises:
            TypeError: When values is not an iterable.

        """
        for value in values:
            try:
                self.add(value)
            except TypeError:
                pass

    def get_errors(self):
        """Return a list of errors contained.

        Returns:
            list(ValidationError): A list of all errors (but not warnings) that are present within the log.

        Todo:
            Add explicit tests.

        """
        return [err for err in self if err.status == 'error']

    def get_errors_or_warnings_by_category(self, err_category):
        """Return a list of errors or warnings of the specified category.

        Args:
            err_category (str): The category of the error to look for.

        Returns:
            list(ValidationError): A list of errors and warnings of the specified category that are present within the log.

        Todo:
            Add explicit tests.

        """
        return [err for err in self._values if err.category == err_category]

    def get_errors_or_warnings_by_name(self, err_name):
        """Return a list of errors or warnings with the specified name.

        Args:
            err_name (str): The name of the error to look for.

        Returns:
            list(ValidationError): A list of errors and warnings with the specified name that are present within the log.

        Todo:
            Add explicit tests.

        """
        return [err for err in self._values if err.name == err_name]

    def get_errors_or_warning_by_type(self, err_type):
        """Return a list of errors or warnings of the specified type.

        Args:
            err_type (type): The type of the error to look for.

        Returns:
            list(ValidationError): A list of errors and warnings of the specified type that are present within the log.

        Todo:
            Add explicit tests.

        """
        return [err for err in self._values if err.base_exception == err_type]

    def get_warnings(self):
        """Return a list of warnings contained.

        Returns:
            list(ValidationError): A list of all warnings (but not errors) that are present within the log.

        Todo:
            Add explicit tests.

        """
        return [err for err in self if err.status == 'warning']


def _check_codes(dataset, codelist):
    """Determine whether a given Dataset has values from the specified Codelist where expected.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Codelist values within.
        codelist (iati.codelists.Codelist): The Codelist to check values from.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    error_log = ValidationErrorLog()
    mappings = iati.default.codelist_mapping()

    for mapping in mappings[codelist.name]:
        base_xpath = mapping['xpath']
        condition = mapping['condition']
        split_xpath = base_xpath.split('/')
        parent_el_xpath = '/'.join(split_xpath[:-1])
        attr_name = split_xpath[-1:][0][1:]

        if condition is None:
            parent_el_xpath = parent_el_xpath + '[@' + attr_name + ']'
        else:
            parent_el_xpath = parent_el_xpath + '[' + condition + ' and @' + attr_name + ']'

        # some nasty string manipulation to make the `//@xml:lang` mapping work
        while not parent_el_xpath.startswith('//'):
            parent_el_xpath = '/' + parent_el_xpath
        if parent_el_xpath.startswith('//['):
            parent_el_xpath = '//*[' + parent_el_xpath[3:]

        parents_to_check = dataset.xml_tree.xpath(parent_el_xpath)

        for parent in parents_to_check:
            code = parent.attrib[attr_name]

            if code not in codelist.codes:
                line_number = parent.sourceline  # used via `locals()` # pylint: disable=unused-variable

                if codelist.complete:
                    error = ValidationError('err-code-not-on-codelist', locals())
                else:
                    error = ValidationError('warn-code-not-on-codelist', locals())

                error.actual_value = code

                error_log.add(error)

    return error_log


def _check_codelist_values(dataset, schema):
    """Check whether a given Dataset has values from Codelists that have been added to a Schema where expected.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Codelist values within.
        schema (iati.schemas.Schema): The Schema to locate Codelists within.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    error_log = ValidationErrorLog()

    for codelist in schema.codelists:
        error_log.extend(_check_codes(dataset, codelist))

    return error_log


def _check_is_iati_xml(dataset, schema):
    """Check whether a given Dataset contains valid IATI XML.

    Args:
        dataset (iati.data.Dataset): The Dataset to check validity of.
        schema (iati.schemas.Schema): The Schema to validate the Dataset against.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    Raises:
        iati.exceptions.SchemaError: An error occurred in the parsing of the Schema.

    Todo:
        Create test against a bad Schema.

    """
    error_log = ValidationErrorLog()

    try:
        validator = schema.validator()
    except iati.exceptions.SchemaError as err:
        raise err

    try:
        validator.assertValid(dataset.xml_tree)
    except etree.DocumentInvalid as doc_invalid:
        for log_entry in doc_invalid.error_log:  # pylint: disable=no-member
            error = _create_error_for_lxml_log_entry(log_entry)
            error_log.add(error)

    return error_log


def _check_is_xml(maybe_xml):
    """Check whether a given parameter is valid XML.

    Args:
        maybe_xml (str): An string that may or may not contain valid XML.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    Todo:
        Consider how a Dataset may be passed when creating errors so that context can be obtained.

    """
    error_log = ValidationErrorLog()

    if isinstance(maybe_xml, iati.data.Dataset):
        maybe_xml = maybe_xml.xml_str

    try:
        parser = etree.XMLParser()
        _ = etree.fromstring(maybe_xml.strip(), parser)
    except etree.XMLSyntaxError:
        for log_entry in parser.error_log:
            error = _create_error_for_lxml_log_entry(log_entry)
            error_log.add(error)
    except (AttributeError, TypeError, ValueError):
        problem_var_type = type(maybe_xml)  # used via `locals()` # pylint: disable=unused-variable
        error = ValidationError('err-not-xml-not-string', locals())
        error_log.add(error)

    # the parser does not cause any errors when given an empty string, so this needs handling separately
    if error_log == ValidationErrorLog() and maybe_xml.strip() == '':
        err_name = 'err-not-xml-empty-document'
        err = 'A file or string containing no data is not XML.'  # used via `locals()` # pylint: disable=unused-variable
        error = ValidationError(err_name, locals())
        error_log.add(error)

    return error_log


def _check_rules(dataset, ruleset):
    """Determine whether a given Dataset conforms with a provided Ruleset.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Ruleset conformance with.
        ruleset (iati.code.Ruleset): The Ruleset to check conformance with.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    error_log = ValidationErrorLog()
    error_found = False

    for rule in ruleset.rules:
        validation_status = rule.is_valid_for(dataset)
        if validation_status is None:
            # A result of `None` signifies that a rule was skipped.
            error = ValidationError('warn-rule-skipped', locals())
            error_log.add(error)
        elif validation_status is False:
            # A result of `False` signifies that a rule did not pass.
            error = _create_error_for_rule(rule)
            error_log.add(error)
            error_found = True

    if error_found:
        # Add a ruleset error if at least one rule error was found.
        error = ValidationError('err-ruleset-conformance-fail', locals())
        error_log.add(error)

    return error_log


def _check_ruleset_conformance(dataset, schema):
    """Check whether a given Dataset conforms with Rulesets that have been added to a Schema.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Ruleset conformance with.
        schema (iati.schemas.Schema): The Schema to locate Rulesets within.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    error_log = ValidationErrorLog()

    for ruleset in schema.rulesets:
        error_log.extend(_check_rules(dataset, ruleset))

    return error_log


def _conforms_with_ruleset(dataset, schema):
    """Determine whether a given Dataset conforms with Rulesets that have been added to a Schema.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Ruleset conformance with.
        schema (iati.schemas.Schema): The Schema to locate Rulesets within.

    Returns:
        bool: A boolean indicating whether the given Dataset conforms with Rulesets attached to the given Schema.

    """
    error_log = _check_ruleset_conformance(dataset, schema)

    return not error_log.contains_errors()


def _correct_codelist_values(dataset, schema):
    """Determine whether a given Dataset has values from Codelists that have been added to a Schema where expected.

    Args:
        dataset (iati.data.Dataset): The Dataset to check Codelist values within.
        schema (iati.schemas.Schema): The Schema to locate Codelists within.

    Returns:
        bool: A boolean indicating whether the given Dataset has values from the specified Codelists where they should be.

    """
    error_log = _check_codelist_values(dataset, schema)

    return not error_log.contains_errors()


def _create_error_for_lxml_log_entry(log_entry):  # pylint: disable=invalid-name
    """Parse a log entry from an lxml error log and convert it to a IATI ValidationError.

    Args:
        log_entry (etree._LogEntry): A log entry from an `etree.XMLSyntaxError` or `etree.DocumentInvalid`.

    Returns:
        ValidationError: An IATI ValidationError that contains the information from the log entry.

    Todo:
        Create a small program to determine the common types of errors so that they can be handled as special cases with detailed help info.

        Determine whether there should be a range of uncategorised errors rather than just 'err-not-xml-uncategorised-xml-syntax-error' eg. IATI error vs. XML error.

    """
    # set the `err` variable so it can be used in error string formatting via locals()
    err = log_entry

    # configure local variables for the creation of the error
    line_number = err.line  # used via `locals()`# pylint: disable=unused-variable
    column_number = err.column  # used via `locals()`# pylint: disable=unused-variable

    # undertake the mapping between error name formats
    lxml_to_iati_error_mapping = {
        'ERR_DOCUMENT_EMPTY': 'err-not-xml-empty-document',
        'ERR_DOCUMENT_END': 'err-not-xml-content-at-end',
        'ERR_INTERNAL_ERROR': 'err-lxml-internal-error',
        'ERR_INVALID_ENCODING': 'err-encoding-invalid',
        'ERR_UNSUPPORTED_ENCODING': 'err-encoding-unsupported',
        'ERR_RESERVED_XML_NAME': 'err-not-xml-xml-text-decl-only-at-doc-start',
        'SCHEMAV_CVC_COMPLEX_TYPE_2_3': 'err-not-iati-xml-non-whitespace-in-element-only',
        'SCHEMAV_CVC_COMPLEX_TYPE_3_2_1': 'err-not-iati-xml-forbidden-attribute',
        'SCHEMAV_CVC_COMPLEX_TYPE_3_2_2': 'err-not-iati-xml-forbidden-attribute',
        'SCHEMAV_CVC_COMPLEX_TYPE_4': 'err-not-iati-xml-missing-attribute',
        'SCHEMAV_CVC_DATATYPE_VALID_1_2_1': 'err-not-iati-xml-incorrect-datatype',
        'SCHEMAV_CVC_ELT_1': 'err-not-iati-xml-root-element-undeclared',
        'SCHEMAV_ELEMENT_CONTENT': 'err-not-iati-xml-missing-required-element'
    }

    try:
        err_name = lxml_to_iati_error_mapping[err.type_name]
    except KeyError:
        err_name = 'err-not-xml-uncategorised-xml-syntax-error'

    error = ValidationError(err_name, locals())

    return error


def _create_error_for_rule(rule):
    """Parse a Rule skip or failure and convert it into an IATI ValidationError.

    Args:
        rule (iati.rulesets.Rule): The Rule which has either skipped or failed.

    Returns:
        ValidationError: An IATI ValidationError that contains information about the Rule that has failed.

    Todo:
        Determine whether there should be a range of uncategorised errors for various ways Ruleset validation may fail, rather than just 'err-rule-uncategorised-conformance-fail'.

    """
    # undertake the mapping between Rule subclass and error name formats
    rule_to_iati_error_mapping = {
        'atleast_one': 'err-rule-at-least-one-conformance-fail',
        'date_order': 'err-rule-date-order-conformance-fail',
        'dependent': 'err-rule-dependent-conformance-fail',
        'no_more_than_one': 'err-rule-no-more-than-one-conformance-fail',
        'regex_matches': 'err-rule-regex-matches-conformance-fail',
        'regex_no_matches': 'err-rule-regex-no-matches-conformance-fail',
        'startswith': 'err-rule-starts-with-conformance-fail',
        'sum': 'err-rule-sum-conformance-fail',
        'unique': 'err-rule-unique-conformance-fail'
    }

    try:
        err_name = rule_to_iati_error_mapping[rule.name]
    except KeyError:
        err_name = 'err-rule-uncategorised-conformance-fail'

    error = ValidationError(err_name, locals())

    return error


def full_validation(dataset, schema):
    """Perform full validation on a Dataset.

    Args:
        dataset (iati.Dataset): The Dataset to check validity of.
        schema (iati.Schema): The Schema to validate the Dataset against.

    Warning:
        Parameters are likely to change in some manner.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    Todo:
        Create test against a bad Schema.

    """
    error_log = ValidationErrorLog()

    error_log.extend(_check_is_xml(dataset))
    error_log.extend(_check_codelist_values(dataset, schema))
    error_log.extend(_check_ruleset_conformance(dataset, schema))

    return error_log


def get_error_codes():
    """Return a dictionary of the possible error codes and their information.

    Returns:
        dict: A dictionary of error codes.

    Raises:
        KeyError: When a specified base_exception is not a valid type of exception.

    Todo:
        Raise the correct error for incorrect base_exception values.
        Raise an error when there is a problem with non-base_exception-related errors.

    """
    err_codes_str = iati.resources.load_as_string(iati.resources.get_lib_data_path('validation_err_codes.yaml'))
    err_codes_list_of_dict = yaml.safe_load(err_codes_str)
    # yaml parses the values into a list of dicts, so they need combining into one
    err_codes_dict = {k: v for code in err_codes_list_of_dict for k, v in code.items()}

    # convert name of exception into reference to the relevant class
    for err in err_codes_dict.values():
        # python2/3 have exceptions in different modules, though six and future do not appear to have a standard workaround for this
        try:
            err['base_exception'] = getattr(sys.modules['builtins'], err['base_exception'])
        except KeyError:
            err['base_exception'] = getattr(sys.modules['exceptions'], err['base_exception'])

    return err_codes_dict


def is_iati_xml(dataset, schema):
    """Determine whether a given Dataset's XML is valid against the specified Schema.

    Args:
        dataset (iati.data.Dataset): The Dataset to check validity of.
        schema (iati.schemas.Schema): The Schema to validate the Dataset against.

    Warning:
        Parameters are likely to change in some manner.

    Returns:
        bool: A boolean indicating whether the given Dataset is valid XML against the given Schema.

    Raises:
        iati.exceptions.SchemaError: An error occurred in the parsing of the Schema.

    Todo:
        Create test against a bad Schema.

    """
    return not _check_is_iati_xml(dataset, schema).contains_errors()


def is_valid(dataset, schema):
    """Determine whether a given Dataset is valid against the specified Schema.

    Args:
        dataset (iati.Dataset): The Dataset to check validity of.
        schema (iati.Schema): The Schema to validate the Dataset against.

    Warning:
        Parameters are likely to change in some manner.

    Returns:
        bool: A boolean indicating whether the given Dataset is valid against the given Schema.

    Todo:
        Create test against a bad Schema.

    """
    try:
        iati_xml = is_iati_xml(dataset, schema)
        if not iati_xml:
            return False
    except iati.exceptions.SchemaError:
        return False

    return _correct_codelist_values(dataset, schema) and _conforms_with_ruleset(dataset, schema)


def is_xml(maybe_xml):
    """Determine whether a given parameter is XML.

    Args:
        maybe_xml (str): An string that may or may not be valid XML.

    Returns:
        bool: A boolean indicating whether the given Dataset is valid XML.

    """
    error_log = _check_is_xml(maybe_xml)

    return not error_log.contains_errors()


def validate_is_iati_xml(dataset, schema):
    """Check whether a Dataset contains valid IATI XML.

    Args:
        dataset (iati.Dataset): The Dataset to check validity of.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    return _check_is_iati_xml(dataset, schema)


def validate_is_xml(maybe_xml):
    """Check whether a Dataset contains valid XML.

    Args:
        maybe_xml (str): An string that may or may not be valid XML.

    Returns:
        iati.validator.ValidationErrorLog: A log of the errors that occurred.

    """
    return _check_is_xml(maybe_xml)



"""A module containing tests for the library representation of IATI constants."""
import iati.constants


class TestConstants(object):
    """A container for tests relating to IATI software constants."""

    def test_namespace(self):
        """Check that the NAMESPACE constant is a string."""
        assert isinstance(iati.constants.NAMESPACE, str)

    def test_nsmap(self):
        """Check that the NSMAP constant is a dictionary with the correct key."""
        assert isinstance(iati.constants.NSMAP, dict)
        assert isinstance(iati.constants.NSMAP['xsd'], str)

    def test_standard_versions_all_are_numbers(self):
        """Check that each item in standard versions is a string that can be considered to be a decimal number."""
        for version in iati.constants.STANDARD_VERSIONS:
            assert isinstance(version, str)
            assert float(version)

    def test_standard_versions_correct_format(self):
        """Check that standard versions is in the correct format."""
        assert isinstance(iati.constants.STANDARD_VERSIONS, list)

    def test_standard_versions_correct_number(self):
        """Check that standard versions has the expected number of items."""
        assert len(iati.constants.STANDARD_VERSIONS) == 4

    def test_standard_versions_major_all_are_integers(self):
        """Check that each major version is an integer."""
        for major_version in iati.constants.STANDARD_VERSIONS_MAJOR:
            assert isinstance(major_version, int)

    def test_standard_versions_major_correct_number(self):
        """Check that the correct number of major versions are detected."""
        assert len(iati.constants.STANDARD_VERSIONS_MAJOR) == 2



"""A module containing a core representation of IATI Codelists."""
import collections
from lxml import etree
import iati.resources
import iati.utilities


class Codelist(object):
    """Representation of a Codelist as defined within the IATI SSOT.

    Attributes:
        complete (bool): Whether the Codelist is complete or not. If complete, attributes making use of this Codelist must only contain values present on the Codelist. If not complete, this is merely strongly advised.
        codes (:obj:`set` of :obj:`iati.Code`): The codes demonstrating the range of values that the Codelist may represent.
        name (str): The name of the Codelist.

    Warning:
        There are currently a large number of attributes that have been taken straight from the XML without being implemented in code. Some of these may change during implementation.

        The `codes` attribute is currently a set. While functionally correct, it may be slightly confusing because the class is a CodeLIST.

    Todo:
        Create a custom class inheriting from set that only allows Codes to be added.

        Implement and document attributes that are not yet implemented and documented.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, name, xml=None):
        """Initialise a Codelist.

        Any Codes contained within the specified XML are added.

        Args:
            name (str): The name of the codelist being initialised.
            xml (str): An XML representation of a codelist.

        Note:
            Instances of a Codelist should remain independent of a particular version of the IATI Standard. Versioning should be handled elsewhere.

        Warning:
            The format of the constructor is likely to change. It needs to be less reliant on the name acting as a UID,  and allow for other attributes to be defined.

        Todo:
            Raise warnings or errors if the Codelist is unable to initialise correctly.

        """
        def parse_from_xml(xml):
            """Parse a Codelist from the XML that defines it.

            Warning:
                In modifying the parameters required for creating an instance of the class, this is likely to move in some manner.

            Todo:
                Define relevant tests and error handling.

                Handle Codelists without description or name elements.

                Better document side-effects.

            """
            tree = iati.utilities.convert_xml_to_tree(xml)

            self.name = tree.attrib['name']
            for code_el in tree.findall('codelist-items/codelist-item'):
                value = code_el.findtext('code')
                name = code_el.findtext('name/narrative') or code_el.findtext('name')

                if (value is None) and (name is None):
                    msg = "The provided Codelist ({0}) has a Code that does not contain a name or value.".format(self.name)
                    iati.utilities.log_warning(msg)

                if value is None:
                    value = ''
                if name is None:
                    name = ''
                self.codes.add(iati.Code(value, name))

            try:
                self.complete = True if tree.attrib['complete'] == '1' else False
            except KeyError:
                pass

        self.complete = None
        self.codes = set()
        self.name = name

        # a number of placeholder attributes that Codelists have, though are not yet implemented
        self._name_prose = None
        self._description = None
        self._language = None
        self._url = None
        self._ref = None
        self._category_codelist = None

        if xml:
            parse_from_xml(xml)

    def __eq__(self, other):
        """Check Codelist equality.

        This allows uniqueness to be correctly defined upon insertion into a set.

        Todo:
            Utilise all attributes as part of the equality process.

        """
        return ((self.name) == (other.name)) and (collections.Counter(self.codes) == collections.Counter(other.codes))

    def __ne__(self, other):
        """Check Codelist inequality.

        Note:
            This is required since there are no implicit relationships when comparing objects. https://docs.python.org/2/reference/datamodel.html#object.__ne__

        """
        return not self == other

    def __hash__(self):
        """Hash the Codelist.

        This allows uniqueness to be correctly defined upon insertion into a set.

        Todo:
            Utilise all attributes as part of the equality process.

        """
        return hash((self.name, tuple(self.codes)))

    @property
    def xsd_restriction(self):
        """Output the Codelist as an XSD simpleType restriction.

        This tree may be used to specify the type of given elements, allowing insertion and validation within a Schema.

        Returns:
            etree.Element: An XSD simpleType representing this Codelist.

        Warning:
            It is planned to change from Schema-based to Data-based Codelist validation. As such, this property may be removed.

            The name attribute of the generated type is not good and needs changing.

            Does not fully hide the lxml internal workings.

        Todo:
            See whether there are only Codelists of a type other than string.

            Improve naming of the type to reduce potential of clashes.

        """
        type_base_el = etree.Element(
            iati.constants.NAMESPACE + 'simpleType',
            name='{0}-type'.format(self.name),
            nsmap=iati.constants.NSMAP
        )
        restriction_base_el = etree.Element(
            iati.constants.NAMESPACE + 'restriction',
            base='xsd:string',
            nsmap=iati.constants.NSMAP
        )

        for code in self.codes:
            restriction_base_el.append(code.xsd_enumeration)

        type_base_el.append(restriction_base_el)

        return type_base_el


class Code(object):
    """Representation of a Code contained within a Codelist.

    Attributes:
        name (str): The name of the code.
        value (str): The value of the code.

    Todo:
        Implement and document attributes that are not yet implemented and documented.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, value=None, name=''):
        """Initialise a Code.

        Args:
            name (str): The name of the code being initialised.
            value (str): The value of the code being initialised.

        Note:
            Instances of a Code should remain independent of a particular version of the IATI Standard. Versioning should be handled elsewhere.

        Warning:
            The format of the constructor is likely to change. It should include mandatory parameters, and allow for other attributes to be defined.

        """
        self.name = name
        self.value = value

        # a number of placeholder attributes that Codelists have, though are not yet implemented
        self._description = None
        self._category = None
        self._url = None
        self._public_database = False
        self._status = None
        self._activation_date = None
        self._withdrawal_date = None

    def __eq__(self, other):
        """Check Code equality.

        This allows uniqueness to be correctly defined upon insertion into a set.

        Todo:
            Utilise all attributes as part of the equality process.

            Test comparison with strings.

        """
        try:
            return ((self.name) == (other.name)) and ((self.value) == (other.value))
        except AttributeError:
            return self.value == other

    def __ne__(self, other):
        """Check Code inequality.

        Note:
            This is required since there are no implicit relationships when comparing objects. https://docs.python.org/2/reference/datamodel.html#object.__ne__

        """
        return not self == other

    def __hash__(self):
        """Hash the Code.

        This allows uniqueness to be correctly defined upon insertion into a set.

        Todo:
            Utilise all attributes as part of the hashing process.

            Be able to deal with checks against both Codes and strings.

        """
        return hash((self.value))

    @property
    def xsd_enumeration(self):
        """Output the Code as an etree enumeration element.

        Returns:
            etree.Element: An XSD enumeration representing this Codelist.

        Warning:
            It is planned to change from Schema-based to Data-based Codelist validation. As such, this property may be removed.

            Does not fully hide the lxml internal workings.

        """
        return etree.Element(
            iati.constants.NAMESPACE + 'enumeration',
            value=self.value,
            nsmap=iati.constants.NSMAP
        )



"""A module containing IATI exceptions.

Warning:
    Proper thought needs to be put into error handling. The currently defined errors do not handle existing use-cases, let alone those that may exist in the future.
"""


class SchemaError(Exception):
    """An error with an IATI Schema.

    Warning:
        Highly likely to change.

        Should inherit from something less generic than `Exception`.
    """

    pass


class ValidationError(ValueError):
    """An error with XML.

    Indicates that the XML does not conform to the IATI standard.

    Warning:
        Highly likely to change.

        This is too general to identify many specific problems.
    """

    def __init__(self, error_log):
        """Initialise a ValidationError."""
        self.error_log = error_log

        super(ValidationError, self).__init__()



"""A module containing a core representation of IATI Rulesets.

Todo:
    Consider how we should handle lxml errors.

    Remove references to `case`.

"""
# no-member errors are due to using `setattr()` # pylint: disable=no-member
import decimal
import json
import re
import sre_constants
from datetime import datetime
import jsonschema
import six
import iati.default
import iati.utilities


_VALID_RULE_TYPES = ["atleast_one", "dependent", "sum", "date_order", "no_more_than_one", "regex_matches", "regex_no_matches", "startswith", "unique"]


def constructor_for_rule_type(rule_type):
    """Locate the constructor for specific Rule types.

    Args:
        rule_type (str): The name of the type of Rule to identify the class for.

    Returns:
        type: A constructor for a class that inherits from Rule.

    Raises:
        KeyError: When a non-permitted `rule_type` is provided.

    """
    possible_rule_types = {
        'atleast_one': RuleAtLeastOne,
        'date_order': RuleDateOrder,
        'dependent': RuleDependent,
        'no_more_than_one': RuleNoMoreThanOne,
        'regex_matches': RuleRegexMatches,
        'regex_no_matches': RuleRegexNoMatches,
        'startswith': RuleStartsWith,
        'sum': RuleSum,
        'unique': RuleUnique
    }

    return possible_rule_types[rule_type]


class Ruleset(object):
    """Representation of a Ruleset as defined within the IATI SSOT."""

    def __init__(self, ruleset_str=None):
        """Initialise a Ruleset.

        Args:
            ruleset_str (str): A string that represents a Ruleset.

        Raises:
            TypeError: When `ruleset_str` is not a string.
            ValueError: When `ruleset_str` does not validate against the Ruleset Schema or cannot be correctly decoded.

        """
        if ruleset_str is None:
            ruleset_str = ''

        try:
            self.ruleset = json.loads(ruleset_str, object_pairs_hook=iati.utilities.dict_raise_on_duplicates)
        except TypeError:
            raise ValueError('Provided Ruleset string is not a string.')
        except ValueError:  # python2/3 - should be json.decoder.JSONDecodeError at python 3.5+
            if ruleset_str.strip() == '':
                self.ruleset = {}
            else:
                raise ValueError('Provided Ruleset string is not valid JSON.')
        self.validate_ruleset()
        self.rules = set()
        self._set_rules()

    def is_valid_for(self, dataset):
        """Validate a Dataset against the Ruleset.

        Args:
            Dataset (iati.Dataset): A Dataset to be checked for validity against the Ruleset.

        Returns:
            bool:
                `True` when the Dataset is valid against the Ruleset.

                `False` when part or all of the Dataset is not valid against the Ruleset.

        Todo:
            Better design how Skips and ValueErrors are treated. The current True/False/Skip/Error thing is a bit clunky.

        """
        for rule in self.rules:
            try:
                if rule.is_valid_for(dataset) is False:
                    return False
            except ValueError:
                return False

        return True

    def validate_ruleset(self):
        """Validate a Ruleset against the Ruleset Schema.

        Raises:
            ValueError: When `ruleset_str` does not validate against the Ruleset Schema.

        """
        try:
            jsonschema.validate(self.ruleset, iati.default.ruleset_schema())
        except jsonschema.ValidationError:
            raise ValueError

    def _set_rules(self):
        """Set the Rules of the Ruleset.

        Extract each case of each Rule from the Ruleset and add to initialised `rules` set.

        """
        for context, rule in self.ruleset.items():
            for rule_type, cases in rule.items():
                for case in cases['cases']:
                    constructor = constructor_for_rule_type(rule_type)
                    new_rule = constructor(context, case)
                    self.rules.add(new_rule)


class Rule(object):
    """Representation of a Rule contained within a Ruleset.

    Acts as a base class for specific types of Rule that actually check the content of the data.

    Attributes:
        name (str): The type of Rule, as specified in a JSON Ruleset.
        context (str): An XPath expression to locate the elements that the Rule is to be checked against.
        case (dict): Specific configuration for this instance of the Rule.

    Todo:
        Determine whether this should be an Abstract Base Class.

    """

    def __init__(self, context, case):
        """Initialise a Rule.

        Args:
            context (str): An XPath expression to locate the elements that the Rule is to be checked against.
            case (dict): Specific configuration for this instance of the Rule.

        Raises:
            TypeError: When a parameter is of an incorrect type.
            ValueError: When a rule_type is not one of the permitted Rule types.

        """
        self.case = case
        self.context = self._validated_context(context)
        self._valid_rule_configuration(case)
        self._set_case_attributes(case)
        self._normalize_xpaths()

    def __str__(self):
        """Return string to state what the Rule is checking."""
        return 'This is a Rule.'

    def _validated_context(self, context):
        """Check that a valid `context` is given for a Rule.

        Args:
            context (str): The XPath expression that selects XML elements that the Rule acts against.

        Returns:
            str: A valid XPath.

        Raises:
            TypeError: When an argument is given that is not a string.
            ValueError: When `context` is an empty string.

        """
        if isinstance(context, six.string_types):
            if context != '':
                return context
            raise ValueError
        raise TypeError

    def _normalize_xpath(self, path):
        """Normalize a single XPath by combining it with `context`.

        Args:
            path (str): An XPath.

        Raises:
            AttributeError: When the `context` isn't set.
            ValueError: When `path` is an empty string.

        Todo:
            Add some logging.
            Re-evaluate this.

        """
        if path == '':
            raise ValueError
        return '/'.join([self.context, path])

    def _normalize_condition(self):
        """Normalize `condition` xpaths."""
        try:
            self.normalized_paths.append(self._normalize_xpath(self.condition))
        except AttributeError:
            pass

    def _normalize_xpaths(self):
        """Normalize xpaths by combining them with `context`.

        Note:
            May be overridden in child class that does not use `paths`.

        """
        self.normalized_paths = [self._normalize_xpath(path) for path in self.paths]
        self._normalize_condition()

    def _valid_rule_configuration(self, case):
        """Check that a configuration being passed into a Rule is valid for the given type of Rule.

        Args:
            case (dict): A dictionary of values, generally parsed as a case from a Ruleset.

        Raises:
            AttributeError: When the Rule name is unset or does not have the required attributes.
            ValueError: When the case is not valid for the type of Rule.

        Note:
            The `name` attribute on the class must be set to a valid rule_type before this function is called.

        """
        try:
            jsonschema.validate(case, self._ruleset_schema_section())
        except jsonschema.ValidationError:
            raise ValueError

    def _set_case_attributes(self, case):
        """Make the required attributes within a case their own attributes in the class.

        Args:
            case (dict): The case to take values from.

        Todo:
            Set non-required properties such as a `condition`.

        """
        required_attributes = self._case_attributes(self._ruleset_schema_section())
        for attrib in required_attributes:
            setattr(self, attrib, case[attrib])

        optional_attributes = self._case_attributes(self._ruleset_schema_section(), False)
        for attrib in optional_attributes:
            try:
                setattr(self, attrib, case[attrib])
            except KeyError:
                pass

    def _case_attributes(self, partial_schema, required=True):
        """Determine the attributes that must be present given the Schema for the Rule type.

        Args:
            partial_schema (dict): The partial JSONSchema to extract attribute names from.
            required (bool): Specifies whether the attributes to be returned should be required or optional according to the Ruleset specification.

        Returns:
            list of str: The names of required or optional attributes.

        """
        if required:
            return [key for key in partial_schema['properties'].keys() if key != 'condition']
        return [key for key in partial_schema['properties'].keys() if key == 'condition']

    def _ruleset_schema_section(self):
        """Locate the section of the Ruleset Schema relevant for the Rule.

        In doing so, makes required properties required.

        Returns:
            dict: A dictionary of the relevant part of the Ruleset Schema, based on the Rule's name.

        Raises:
            AttributeError: When the Rule name is unset or does not have the required attributes.

        """
        ruleset_schema = iati.default.ruleset_schema()
        partial_schema = ruleset_schema['patternProperties']['.+']['properties'][self.name]['properties']['cases']['items']  # pylint: disable=E1101
        # make all attributes other than 'condition' in the partial schema required
        partial_schema['required'] = self._case_attributes(partial_schema)
        # ensure that the 'paths' array is not empty
        if 'paths' in partial_schema['properties'].keys():
            partial_schema['properties']['paths']['minItems'] = 1

        return partial_schema

    def _find_context_elements(self, dataset):
        """Find the specific elements in context for the Rule.

        Args:
            dataset (iati.Dataset): The Dataset to be chacked for validity against the Rule.

        Returns:
            list of elements: Results of XPath query.

        Raises:
            AttributeError: When an argument is given that does not have the required attributes.

        """
        return dataset.xml_tree.xpath(self.context)

    def _extract_text_from_element_or_attribute(self, context, path):
        """Return a list of strings regardless of whether XPath result is an attribute or an element.

        Args:
            context (etree._Element): An xml Element.
            path (str): An XPath query string.

        Returns:
            list of str: Text values from XPath query results.

        Note:
            `Element.text` will return `None` if it contains no text. This is bad. As such, this is converted to an empty string to prevent TypeErrors.
            `path` should be validated outside of this function to avoid unexpected errors.

        """
        xpath_results = context.xpath(path)
        results = [result if isinstance(result, six.string_types) else result.text for result in xpath_results]
        return ['' if result is None else result for result in results]

    def _condition_met_for(self, context_element):
        """Check for condtions of a given case.

        Args:
            dataset (iati.Dataset): The Dataset to be checked for validity against a Rule.

        Returns:
            bool: Returns `False` when condition not met.
                  Returns `True` when condition is met.
            None: Returns `None` when condition met.

        Warning:
            Current implementation may be vulnerable to XPath injection vulnerabilities.

        Todo:
            Need to assess the possibility of risk and potential counter-measures/avoidance strategies if needed.
            Need to decide whether the implementation of this in Rules should `return None` or `continue`.
            Rename function to sound more truthy.

        """
        try:
            if context_element.xpath(self.condition):
                return True
        except AttributeError:
            return False

        return False

    def is_valid_for(self, dataset):
        """Check whether a Dataset is valid against the Rule.

        Args:
            dataset (iati.Dataset): The Dataset to be checked for validity against the Rule.

        Returns:
            bool or None:
                `True` when the Dataset is valid against the Rule.

                `False` when the Dataset is not valid against the Rule.

                `None` when a condition is met to skip validation.

        Raises:
            TypeError: When a Dataset is not given as an argument.
            ValueError: When a check encounters a completely incorrect value that it is unable to recover from within the definition of the Rule.

        Note:
            May be overridden in child class that does not have the same return structure for boolean results.

        Todo:
            Better design how Skips and ValueErrors are treated. The current True/False/Skip/Error thing is a bit clunky.

        """
        try:
            context_elements = self._find_context_elements(dataset)
        except AttributeError:
            raise TypeError

        if context_elements == list():
            return None

        for context_element in context_elements:
            if self._condition_met_for(context_element):
                return None

            rule_check_result = self._check_against_Rule(context_element)
            if rule_check_result is False:
                return False
            elif rule_check_result is None:
                return None

        return True


class RuleAtLeastOne(Rule):
    """Representation of a Rule that checks that there is at least one Element matching a given XPath.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.

    """

    def __init__(self, context, case):
        """Initialise an `atleast_one` rule."""
        self.name = 'atleast_one'

        super(RuleAtLeastOne, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleAtLeastOne is checking."""
        if len(self.paths) == 1:
            return '`{self.paths[0]}` must be present within each `{self.context}`.'.format(**locals())
        return 'At least one of `{0}` must be present within each `{self.context}`.'.format('` or `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Check `context_element` has at least one specified Element or Attribute.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `False` when the case is found in the Dataset.
                  Return `True` when the case is not found in the Dataset.

        """
        for path in self.paths:
            if context_element.xpath(path):
                return False
        return True

    def is_valid_for(self, dataset):
        """Check whether a Dataset is valid against the Rule.

        Args:
            dataset (iati.Dataset): The Dataset to be checked for validity against the Rule.

        Returns:
            bool or None:
                `True` when the Dataset is valid against the Rule.

                `False` when the Dataset is not valid against the Rule.

                `None` when a condition is met to skip validation.

        Raises:
            TypeError: When a Dataset is not given as an argument.

        """
        parent = super(RuleAtLeastOne, self).is_valid_for(dataset)

        if parent is True:
            return False
        elif parent is None:
            return None
        return True


class RuleDateOrder(Rule):
    """Representation of a Rule that checks that the date value of `more` is the most recent value in comparison to the date value of `less`.

    Attributes:
        less (str): An XPath expression to locate the element containing the date that should be in the past.
        more (str): An XPath expression to locate the element containing the date that should be in the future.
        special_case (str): A value that will be treated as the present when provided as the `less` or `more` value.

    """

    def __init__(self, context, case):
        """Initialise a `date_order` rule."""
        self.name = 'date_order'
        self.special_case = 'NOW'  # Was a constant sort of

        super(RuleDateOrder, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleDateOrder is checking."""
        if self.less == self.special_case and self.more == self.special_case:
            unformatted_str = '`{self.less}` must be chronologically before `{self.more}`. Try working that one out.'
        elif self.less == self.special_case:
            unformatted_str = '`{self.more}` must be in the future within each `{self.context}`.'
        elif self.more == self.special_case:
            unformatted_str = '`{self.less}` must be in the past within each `{self.context}`.'
        else:
            unformatted_str = '`{self.less}` must be chronologically before `{self.more}` within each `{self.context}`.'

        return unformatted_str.format(**locals())

    def _normalize_xpaths(self):
        """Normalize xpaths by combining them with `context`."""
        self.normalized_paths = list()
        if self.less is not self.special_case:
            self.normalized_paths.append(self._normalize_xpath(self.less))

        if self.more is not self.special_case:
            self.normalized_paths.append(self._normalize_xpath(self.more))

        self._normalize_condition()

    def _get_date(self, context_element, path):
        """Retrieve datetime object from an XPath string.

        Args:
            context_element (etree._Element): An XML Element.
            path: (an XPath): The ultimate XPath query to find the desired elements.

        Returns:
            datetime.datetime: A datetime object.

        Raises:
            ValueError:
                When a non-permitted number of unique dates are given for a `less` or `more` value.
                When datetime cannot convert a string of non-permitted characters.
                When non-permitted trailing characters are found after the core date string characters.

        Note:
            Though technically permitted, any dates with a leading '-' character are almost certainly incorrect and are therefore treated as data errors.

        Todo:
            Consider breaking this function down further.

        """
        if path == self.special_case:
            return datetime.today()

        dates = self._extract_text_from_element_or_attribute(context_element, path)
        if dates == list() or not dates[0]:
            return
        # Checks that anything after the YYYY-MM-DD string is a permitted timezone character
        pattern = re.compile(r'^([+-]([01][0-9]|2[0-3]):([0-5][0-9])|Z)?$')
        if (len(set(dates)) == 1) and pattern.match(dates[0][10:]):
            if len(dates[0]) < 10:
                # '%d' and '%m' are documented as requiring zero-padded dates.as input. This is actually for output. As such, a separate length check is required to ensure zero-padded values.
                raise ValueError
            return datetime.strptime(dates[0][:10], '%Y-%m-%d')
        raise ValueError

    def _check_against_Rule(self, context_element):
        """Assert that the date value of `less` is chronologically before the date value of `more`.

        Args:
            context_element (etree._Element): An XML Element.

        Return:
            bool: Return `True` when `less` is chronologically before `more`.
                  Return `False` when `less` is not chronologically before `more`.
            None: When a condition is met to skip validation.

        Raises:
            ValueError: When a date is given that is not in the correct xsd:date format.

        Note:
            `date` restricted to 10 characters in order to exclude possible timezone values.

        """
        early_date = self._get_date(context_element, self.less)
        later_date = self._get_date(context_element, self.more)

        try:
            # python2 allows `bool`s to be compared to `None` without raising a TypeError, while python3 does not
            if early_date is None or later_date is None:
                return None

            if early_date > later_date:
                return False
        except TypeError:
            # a TypeError is raised in python3 if either of the dates is None
            return None
        return True


class RuleDependent(Rule):
    """Representation of a Rule that checks that if one of the Elements or Attributes in a given `path` exists then all its dependent Elements or Attributes must also exist.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.

    """

    def __init__(self, context, case):
        """Initialise a `dependent` rule."""
        self.name = 'dependent'

        super(RuleDependent, self).__init__(context, case)

    def __str__(self):
        """Return string stating what TestRuleDependent is checking."""
        if len(self.paths) == 1:
            return 'Within each `{self.context}`, either `{self.paths[0]}` exists or it does not. As such, this Rule is always True.'.format(**locals())
        return 'Within each `{self.context}`, either none of `{0}` must exist, or they must all exist.'.format('` or `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Assert that either all given `paths` or none of the given `paths` exist for the `context_element`.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when all dependent `paths` are found in the Dataset, if any exist.
                  Return `False` when only some of the dependent `paths` are found in the Dataset.

        """
        unique_paths = set(self.paths)
        found_paths = 0
        for path in unique_paths:
            results = context_element.xpath(path)
            if results != list():
                found_paths += 1

        if found_paths not in [0, len(unique_paths)]:
            return False
        return True


class RuleNoMoreThanOne(Rule):
    """Representation of a Rule that checks that there is no more than one Element or Attribute matching a given XPath.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.

    """

    def __init__(self, context, case):
        """Initialise a `no_more_than_one` rule."""
        self.name = 'no_more_than_one'

        super(RuleNoMoreThanOne, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleNoMoreThanOne is checking."""
        if len(self.paths) == 1:
            return '`{self.paths[0]}` must occur zero or one times within each `{self.context}`.'.format(**locals())
        return 'There must be no more than one element or attribute matched at `{0}` within each `{self.context}`.'.format('` or `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Check `context_element` has no more than one result for a specified Element or Attribute.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when one result or no results are found in the Dataset.
                  Return `False` when more than one result is found in the Dataset.

        """
        unique_paths = set(self.paths)

        found_elements = 0

        for path in unique_paths:
            results = context_element.xpath(path)
            found_elements += len(results)

        if found_elements > 1:
            return False
        return True


class RuleRegexMatches(Rule):
    """Representation of a Rule that checks that the text of the given paths must match the regex value.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.
        regex (str): A Perl-style regular expression.

    """

    def __init__(self, context, case):
        """Initialise a `regex_matches` Rule.

        Raises:
            ValueError: When the case does not contain valid regex.

        """
        self.name = 'regex_matches'

        super(RuleRegexMatches, self).__init__(context, case)

        if self.regex == '':
            raise ValueError
        try:
            re.compile(self.regex)
        except sre_constants.error:
            raise ValueError

    def __str__(self):
        """Return string stating what RuleRegexMatches is checking."""
        if len(self.paths) == 1:
            return 'Each `{self.paths[0]}` within each `{self.context}` must match the regular expression `{self.regex}`.'.format(**locals())
        return 'Each instance of `{0}` within each `{self.context}` must match the regular expression `{self.regex}`.'.format('` and `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Assert that the text of the given `paths` matches the regex value.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when the given `path` text matches the given regex.
                  Return `False` when the given `path` text does not match the given regex.

        """
        pattern = re.compile(self.regex)

        for path in self.paths:
            strings_to_check = self._extract_text_from_element_or_attribute(context_element, path)
            for string_to_check in strings_to_check:
                if not pattern.search(string_to_check):
                    return False
        return True


class RuleRegexNoMatches(Rule):
    """Representation of a Rule that checks that the text of the given `paths` must not match the regex value.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.
        regex (str): A Perl-style regular expression.

    """

    def __init__(self, context, case):
        """Initialise a `regex_no_matches` Rule.

        Raises:
            ValueError: When the case does not contain valid regex.

        """
        self.name = 'regex_no_matches'

        super(RuleRegexNoMatches, self).__init__(context, case)

        if self.regex == '':
            raise ValueError
        try:
            re.compile(self.regex)
        except sre_constants.error:
            raise ValueError

    def __str__(self):
        """Return string stating what RuleRegexNoMatches is checking."""
        if len(self.paths) == 1:
            return 'Each `{self.paths[0]}` within each `{self.context}` must not match the regular expression `{self.regex}`.'.format(**locals())
        return 'Each instance of `{0}` within each `{self.context}` must not match the regular expression `{self.regex}`.'.format('` and `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Assert that no text of the given `paths` matches the regex value.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when the given `path` text does not match the given regex.
                  Return `False` when the given `path` text matches the given regex.

        """
        pattern = re.compile(self.regex)

        for path in self.paths:
            strings_to_check = self._extract_text_from_element_or_attribute(context_element, path)
            for string_to_check in strings_to_check:
                if pattern.search(string_to_check):
                    return False
        return True


class RuleStartsWith(Rule):
    """Representation of a Rule that checks that the prefixing text of each text value for `path` matches the `start` text value.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.
        start (str): An XPath expression to locate a single element. The text of this element is used as the prefix value for the Rule.

    """

    def __init__(self, context, case):
        """Initialise a `startswith` Rule."""
        self.name = 'startswith'

        super(RuleStartsWith, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleStartsWith is checking."""
        if len(self.paths) == 1:
            return 'Each `{self.paths[0]}` within each `{self.context}` must start with the value present at `{self.start}`.'.format(**locals())
        return 'Each instance of `{0}` within each `{self.context}` must start with the value present at `{self.start}`.'.format('` and `'.join(self.paths), **locals())

    def _normalize_xpaths(self):
        """Normalize xpaths by combining them with `context`."""
        super(RuleStartsWith, self)._normalize_xpaths()

        self.normalized_paths.append(self._normalize_xpath(self.start))

    def _check_against_Rule(self, context_element):
        """Assert that the prefixing text of all given `paths` starts with the text of `start`.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when the `path` text starts with the text value of `start`.
                  Return `False` when the `path` text does not start with the text value of `start`.

        Raises:
            ValueError: When more than one element or attribute is retured for the prefix value.
                        When no results are returned for the prefix value.

        """
        start_results = self._extract_text_from_element_or_attribute(context_element, self.start)

        if len(start_results) > 1:
            raise ValueError
        try:
            prefix = start_results[0]
        except IndexError:
            raise ValueError

        for path in self.paths:
            strings_to_check = self._extract_text_from_element_or_attribute(context_element, path)
            for string_to_check in strings_to_check:
                if not string_to_check.startswith(prefix):
                    return False
        return True


class RuleSum(Rule):
    """Representation of a Rule that checks that the values in given `path` attributes must sum to the given `sum` value.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.
        sum (float): The value that the contents of the located elements and attributes must sum to.

    """

    def __init__(self, context, case):
        """Initialise a `sum` rule."""
        self.name = 'sum'

        super(RuleSum, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleSum is checking."""
        return 'Within each `{self.context}`, the sum of values matched at `{0}` must be `{self.sum}`.'.format('` and `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Assert that the total of the values given in `paths` match the given `sum` value.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when the `path` values total to the `sum` value.
                  Return `False` when the `path` values do not total to the `sum` value.
            None: When no elements are found for the specified `paths`.

        Raises:
            ValueError: When the `path` value is not numeric.

        """
        unique_paths = set(self.paths)
        values_in_context = list()

        for path in unique_paths:
            values_to_sum = self._extract_text_from_element_or_attribute(context_element, path)
            for value in values_to_sum:
                try:
                    values_in_context.append(decimal.Decimal(value))
                except decimal.InvalidOperation:
                    raise ValueError

        if values_in_context == list():
            return None

        if sum(values_in_context) != decimal.Decimal(str(self.sum)):
            return False
        return True


class RuleUnique(Rule):
    """Representation of a Rule that checks that the text of each given path must be unique.

    Attributes:
        paths (list of str): A list of XPath expressions. These are evaluated to locate the elements that the Rule is to operate on.

    """

    def __init__(self, context, case):
        """Initialise a `unique` rule."""
        self.name = 'unique'

        super(RuleUnique, self).__init__(context, case)

    def __str__(self):
        """Return string stating what RuleUnique is checking."""
        return 'Within each `{self.context}`, the text contained within each of the elements and attributes matched by `{0}` must be unique.'.format('` and `'.join(self.paths), **locals())

    def _check_against_Rule(self, context_element):
        """Assert that the given `paths` are not found for `context_element` more than once.

        Args:
            context_element (etree._Element): An XML Element.

        Returns:
            bool: Return `True` when repeated text is not found in the Dataset.
                  Return `False` when repeated text is found in the Dataset.

        Todo:
            Consider better methods for specifying which elements in the tree contain non-permitted duplication, such as bucket sort.

        """
        unique_paths = set(self.paths)
        all_content = list()
        unique_content = set()

        for path in unique_paths:
            strings_to_check = self._extract_text_from_element_or_attribute(context_element, path)
            for string_to_check in strings_to_check:
                all_content.append(string_to_check)
                unique_content.add(string_to_check)

        if len(all_content) != len(unique_content):
            return False
        return True



"""Configuration to exist in the global scope for pytest."""
import collections
import pytest
import iati.default
import iati.resources


@pytest.fixture(params=[
    ('2.02', 62),  # There are 38 embedded codelists at v2.02, plus 24 non-embedded codelists (which are valid for any version)
    ('2.01', 61),  # There are 37 embedded codelists at v2.01, plus 24 non-embedded codelists (which are valid for any version)
    ('1.05', 59),  # There are 35 embedded codelists at v1.05, plus 24 non-embedded codelists (which are valid for any version)
    ('1.04', 59)  # There are 35 embedded codelists at v1.04, plus 24 non-embedded codelists (which are valid for any version)
])
def codelist_lengths_by_version(request):
    """Return a tuple containing versions of the Standard, and the number of Codelists for that version.

    Format: `(version=[standardVersion], expected_length=[numCodelists])`

    """
    output = collections.namedtuple('output', 'version expected_length')
    return output(version=request.param[0], expected_length=request.param[1])


@pytest.fixture
def schema_ruleset():
    """Return a schema with the Standard Ruleset added.

    Returns:
        A valid Activity Schema with the Standard Ruleset added.

    """
    schema = iati.default.activity_schema(None, False)
    ruleset = iati.default.ruleset()

    schema.rulesets.add(ruleset)

    return schema


@pytest.fixture(params=iati.constants.STANDARD_VERSIONS)
def standard_version_mandatory(request):
    """Return a list that can be passed to a function using the argument list unpacking functionality.

    For more information about unpacking argument lists, see https://docs.python.org/3.6/tutorial/controlflow.html#unpacking-argument-lists

    Example:
        The returned list can be used to test functions (such as `iati.default.codelists`) which has an optional parameter for the version, or can expect `version=None`. It has an optional parameter after the version.
        In this case test usage would be `iati.default.codelists(*standard_version_mandatory)`.

    Returns:
        list: A string which corresponds to a version of the Standard.

    """
    return [request.param]


@pytest.fixture(params=['no_arguments', None] + iati.constants.STANDARD_VERSIONS)
def standard_version_optional(request):
    """Return a list that can be passed to a function using the argument list unpacking functionality.

    For mor information about unpacking argument lists, see https://docs.python.org/3.6/tutorial/controlflow.html#unpacking-argument-lists

    Example:
        The returned list can be used to test functions (such as `get_all_codelist_paths`) which has an optional parameter for the version, or can expect `version=None`.,
        In this case test usage would be `get_all_codelist_paths(*standard_version_optional)`.

    Returns:
        list: Either i) an empty list, ii) a list containing None, or iii) a string which corresponds to a version of the Standard.

    """
    if request.param == 'no_arguments':
        return []
    return [request.param]



"""A module containing utility functions."""
import logging
import os
from io import StringIO
from lxml import etree
import iati.constants


def add_namespace(tree, new_ns_name, new_ns_uri):
    """Add a namespace to a Schema.

    Args:
        tree (etree._ElementTree): The ElementTree to add a namespace to.

        new_ns_name (str): The name of the new namespace. Must be valid against https://www.w3.org/TR/REC-xml-names/#NT-NSAttName

        new_ns_uri (str): The URI for the new namespace. Must be non-empty and valid against https://www.ietf.org/rfc/rfc2396.txt

    Returns:
        etree.ElementTree: A copy of the provided `tree`, modified to include the specified namespace.

    Raises:
        TypeError: If an attempt is made to add a namespace to something other than a ElementTree.
        ValueError: If the namespace name or URI are invalid values.
        ValueError: If the namespace name already exists.

    Note:
        lxml does not allow modification of namespaces within a tree that already exists. As such, string manipulation is used. https://bugs.launchpad.net/lxml/+bug/555602

    Todo:
        Also add new namespaces to Datasets.

        Add checks for the format of new_ns_name - for syntax, see: https://www.w3.org/TR/REC-xml-names/#NT-NSAttName

        Add checks for the format of new_ns_uri - for syntax, see: https://www.ietf.org/rfc/rfc2396.txt

        Tidy this up.

    """
    if not isinstance(tree, etree._ElementTree):  # pylint: disable=protected-access
        msg = "The `tree` parameter must be of type `etree._ElementTree` - it was of type {0}".format(type(tree))
        iati.utilities.log_error(msg)
        raise TypeError(msg)
    if not isinstance(new_ns_name, str) or not new_ns_name:
        msg = "The `new_ns_name` parameter must be a non-empty string."
        iati.utilities.log_error(msg)
        raise ValueError(msg)
    if not isinstance(new_ns_uri, str) or not new_ns_uri:
        msg = "The `new_ns_uri` parameter must be a valid URI."
        iati.utilities.log_error(msg)
        raise ValueError(msg)

    initial_nsmap = tree.getroot().nsmap
    # prevent modification of existing namespaces
    if new_ns_name in initial_nsmap:
        if new_ns_uri == initial_nsmap[new_ns_name]:
            return tree
        else:
            msg = "There is already a namespace called {0}.".format(new_ns_name)
            iati.utilities.log_error(msg)
            raise ValueError(msg)

    # to add new namespace, use algorithm from http://stackoverflow.com/a/11350061
    schema_str = etree.tostring(tree.getroot(), pretty_print=True).decode('unicode_escape')
    interim_tree = etree.ElementTree(element=None, file=StringIO(schema_str))
    root = interim_tree.getroot()
    nsmap = root.nsmap
    nsmap[new_ns_name] = new_ns_uri
    new_root = etree.Element(root.tag, nsmap=nsmap)
    new_root[:] = root[:]
    new_tree = etree.ElementTree(new_root)

    return new_tree


def convert_tree_to_schema(tree):
    """Convert an etree to a Schema.

    Args:
        tree (etree._ElementTree): An XML element tree representing an XML Schema.

    Returns:
        etree.XMLSchema: The XML schema that the provided tree represented.

    Warning:
        Should raise exceptions when there are errors during execution.

        Needs to better distinguish between an `etree.XMLSchema`, an `iati.Schema`, an `iati.ActivitySchema` and an `iati.OrganisationSchema`.

        Does not fully hide the lxml internal workings.

    Todo:
        Surround schema conversion with error handling.

    """
    return etree.XMLSchema(tree)


def convert_xml_to_tree(xml):
    """Convert an XML string into an etree.

    Args:
        xml (str): An XML string to be converted.

    Returns:
        etree._Element: An lxml element tree representing the provided XML.

    Warning:
        Does not fully hide the lxml internal workings.

    Raises:
        ValueError: The XML provided was something other than a string.
        lxml.etree.XMLSyntaxError: There was an error with the syntax of the provided XML.

    """
    try:
        tree = etree.fromstring(xml)
        return tree
    except etree.XMLSyntaxError as xml_syntax_err:
        msg = "There was a problem with the provided XML, and it could therefore not be turned into a tree."
        iati.utilities.log_error(msg)
        raise xml_syntax_err
    except ValueError:
        msg = "To parse XML into a tree, the XML must be a string, not a {0}.".format(type(xml))
        iati.utilities.log_error(msg)
        raise ValueError(msg)


def dict_raise_on_duplicates(ordered_pairs):
    """Reject duplicate keys in a dictionary.

    RFC4627 merely says that keys in a JSON file SHOULD be unique. As such, `json.loads()` permits duplicate keys, and overwrites earlier values with those later in the string.

    In creating Rulesets, we wish to forbid duplicate keys. As such, this function may be used to do this.

    Algorithm from https://stackoverflow.com/a/14902564

    Args:
        ordered_pairs (list(tuple)): A list of (key, value) pairs.

    Raises:
        ValueError: When there are duplicate keys.

    Returns:
        dict: A dictionary constructed from `ordered_pairs`.

    """
    duplicate_free_dict = {}
    for key, value in ordered_pairs:
        if key in duplicate_free_dict:
            raise ValueError("duplicate key: %r" % (key,))
        else:
            duplicate_free_dict[key] = value
    return duplicate_free_dict


def log(lvl, msg, *args, **kwargs):
    """Log a message of some level.

    Args:
        lvl (int): The level of message being logged.
        msg (str): The message that is to be logged.
        *args
        **kwargs

    Warning:
        Potentially too tightly coupled to the Python `logging` module.

        Logging needs to be defined in a much more useful and configurable manner.

        Logging should not fill up logfiles at lightspeed unless this is specifically desired.

        Outputs should be more easily parsable.

    """
    logging.basicConfig(
        filename=os.path.join(iati.constants.LOG_FILE_NAME),
        format='%(asctime)s %(levelname)s:%(name)s: %(message)s %(stack_info)s',
        level=logging.DEBUG
    )
    logger = logging.getLogger(iati.constants.LOGGER_NAME)
    logger.log(lvl, msg, *args, **kwargs)


def log_error(msg, *args, **kwargs):
    """Log an error.

    Args:
        msg (str): The message that is to be logged.
        *args
        **kwargs

    Warning:
        Potentially too tightly coupled to the Python `logging` module.

    """
    log(logging.ERROR, msg, *args, **kwargs)


def log_exception(msg, *args, **kwargs):
    """Log an exception.

    An exception is like an error, but with a stack trace.

    Args:
        msg (str): The message that is to be logged.
        *args
        **kwargs

    Warning:
        Potentially too tightly coupled to the Python `logging` module.

    """
    log(logging.ERROR, msg, exc_info=True, *args, **kwargs)


def log_warning(msg, *args, **kwargs):
    """Log a warning.

    Args:
        msg (str): The message that is to be logged.
        *args
        **kwargs

    Warning:
        Potentially too tightly coupled to the Python `logging` module.

    """
    log(logging.WARN, msg, *args, **kwargs)


def versions_for_integer(integer):
    """Return a list containing the supported versions for the input integer version.

    Args:
        integer (int): The integer version to find the supported version for.

    Returns:
        list of str: Containing the supported versions for the input integer.

    """
    output = list()
    for version in iati.constants.STANDARD_VERSIONS:
        if version.startswith(str(integer)):
            output.append(version)

    return output



"""A module containing tests for the library representation of Codelists."""
import pytest
from lxml import etree
import iati.codelists


class TestCodelistsNonClass(object):
    """Test codelists functionality that is not contained within a class.

    Note:
        There was once functionality regarding mapping files here. That was removed.
    """

    pass


class TestCodelists(object):
    """A container for tests relating to Codelists."""

    @pytest.fixture
    def name_to_set(self):
        """Set a name to give Codelists.

        Returns:
            str: Something that can be provided as a name to Codelists.

        """
        return "test Codelist name"

    def test_codelist_default_attributes(self):
        """Check a Codelist's default attributes are correct."""
        with pytest.raises(TypeError) as excinfo:
            iati.Codelist()  # pylint: disable=E1120

        assert ('__init__() missing 1 required positional argument' in str(excinfo.value)) or ('__init__() takes at least 2 arguments' in str(excinfo.value))

    def test_codelist_name_instance(self, name_to_set):
        """Check a Codelist's attributes are correct when defined with only a name."""
        codelist = iati.Codelist(name_to_set)

        assert set() == codelist.codes
        assert codelist.name == name_to_set

    def test_codelist_add_code(self, name_to_set):
        """Check a Code can be added to a Codelist."""
        codelist = iati.Codelist(name_to_set)
        code = iati.Code()
        codelist.codes.add(code)

        num_codes = len(codelist.codes)

        assert num_codes == 1

    @pytest.mark.xfail
    def test_codelist_add_code_decline_non_code(self, name_to_set):
        """Check something that is not a Code cannot be added to a Codelist."""
        codelist = iati.Codelist(name_to_set)
        not_a_code = True
        codelist.codes.add(not_a_code)

        num_codes = len(codelist.codes)

        assert num_codes == 0

    def test_codelist_define_from_xml(self, name_to_set):
        """Check that a Codelist can be generated from an XML codelist definition."""
        path = iati.resources.get_codelist_path('FlowType')
        xml_str = iati.resources.load_as_string(path)
        codelist = iati.Codelist(name_to_set, xml=xml_str)

        code_names = ['ODA', 'OOF', 'Private grants', 'Private Market', 'Non flow', 'Other flows']
        code_values = ['10', '20', '30', '35', '40', '50']

        assert codelist.name == 'FlowType'
        assert len(codelist.codes) == 6
        for code in codelist.codes:
            assert code.name in code_names
            assert code.value in code_values

    def test_codelist_complete(self):
        """Check that a Codelist can be generated from an XML codelist definition."""
        codelist_name = 'BudgetType'
        path = iati.resources.get_codelist_path(codelist_name)
        xml_str = iati.resources.load_as_string(path)
        codelist = iati.Codelist(codelist_name, xml=xml_str)

        assert codelist.complete is True

    def test_codelist_incomplete(self):
        """Check that a Codelist can be generated from an XML codelist definition."""
        codelist_name = 'Country'
        path = iati.resources.get_codelist_path(codelist_name)
        xml_str = iati.resources.load_as_string(path)
        codelist = iati.Codelist(codelist_name, xml=xml_str)

        assert codelist.complete is False

    def test_codelist_type_xsd(self, name_to_set):
        """Check that a Codelist can turn itself into a type to use for validation."""
        code_value_to_set = "test Code value"
        codelist = iati.Codelist(name_to_set)
        code = iati.Code(code_value_to_set)
        codelist.codes.add(code)

        type_tree = codelist.xsd_restriction

        assert isinstance(type_tree, etree._Element)  # pylint: disable=protected-access
        assert type_tree.tag == iati.constants.NAMESPACE + 'simpleType'
        assert type_tree.attrib['name'] == name_to_set + '-type'
        assert type_tree.nsmap == iati.constants.NSMAP
        assert len(type_tree) == 1

        assert type_tree[0].tag == iati.constants.NAMESPACE + 'restriction'
        assert type_tree[0].nsmap == iati.constants.NSMAP
        assert len(type_tree[0]) == 1

        assert type_tree[0][0].tag == iati.constants.NAMESPACE + 'enumeration'
        assert type_tree[0][0].attrib['value'] == code_value_to_set
        assert type_tree[0][0].nsmap == iati.constants.NSMAP


class TestCodes(object):
    """A container for tests relating to Codes."""

    def test_code_default_attributes(self):
        """Check a Code's default attributes are correct."""
        code = iati.Code()

        assert code.name == ''
        assert code.value is None

    def test_code_value_instance(self):
        """Check a Code's attributes are correct when being defined with only a value."""
        value_to_set = "test Code value"
        code = iati.Code(value_to_set)

        assert code.name == ''
        assert code.value == value_to_set

    def test_code_value_and_name_instance(self):
        """Check a Code's attributes are correct when being defined with a value and name."""
        value_to_set = "test Code value"
        name_to_set = "test Code name"
        code = iati.Code(value_to_set, name_to_set)

        assert code.name == name_to_set
        assert code.value == value_to_set

    def test_code_enumeration_element(self):
        """Check that a Code correctly outputs an enumeration element.

        Todo:
            Test enumerating a Code with no value.

        """
        value_to_set = "test Code value"
        code = iati.Code(value_to_set)

        enum_el = code.xsd_enumeration

        assert isinstance(enum_el, etree._Element)  # pylint: disable=protected-access
        assert enum_el.tag == iati.constants.NAMESPACE + 'enumeration'
        assert enum_el.attrib['value'] == value_to_set
        assert enum_el.nsmap == iati.constants.NSMAP



"""A module containing tests for the library implementation of accessing utilities."""
from lxml import etree
import pytest
import iati.resources
import iati.tests.utilities
import iati.utilities


class TestUtilities(object):
    """A container for tests relating to utilities."""

    @pytest.fixture
    def schema_base_tree(self):
        """Return schema_base_tree."""
        activity_schema_path = iati.resources.get_schema_path(
            iati.tests.utilities.SCHEMA_ACTIVITY_NAME_VALID
        )
        return iati.ActivitySchema(activity_schema_path)._schema_base_tree  # pylint: disable=protected-access

    def test_add_namespace_schema_new(self, schema_base_tree):
        """Check that an additional namespace can be added to a Schema.

        Todo:
            Deal with required changes to iati.constants.NSMAP.

            Add a similar test for Datasets.

        """
        initial_nsmap = schema_base_tree.getroot().nsmap
        ns_name = 'xi'
        ns_uri = 'http://www.w3.org/2001/XInclude'

        tree = iati.utilities.add_namespace(schema_base_tree, ns_name, ns_uri)
        new_nsmap = tree.getroot().nsmap

        assert len(new_nsmap) == len(initial_nsmap) + 1
        assert ns_name not in initial_nsmap
        assert ns_name in new_nsmap
        assert new_nsmap[ns_name] == ns_uri

    def test_add_namespace_schema_already_present(self, schema_base_tree):
        """Check that attempting to add a namespace that already exists changes nothing if the new URI is the same.

        Todo:
            Add a similar test for Datasets.

        """
        initial_nsmap = schema_base_tree.getroot().nsmap
        ns_name = 'xsd'
        ns_uri = 'http://www.w3.org/2001/XMLSchema'

        tree = iati.utilities.add_namespace(schema_base_tree, ns_name, ns_uri)
        new_nsmap = tree.getroot().nsmap

        assert len(new_nsmap) == len(initial_nsmap)
        assert ns_name in initial_nsmap
        assert ns_name in new_nsmap
        assert initial_nsmap[ns_name] == ns_uri
        assert new_nsmap[ns_name] == ns_uri

    def test_add_namespace_schema_already_present_diff_value(self, schema_base_tree):
        """Check that attempting to add a namespace that already exists to a Schema raises an error rather than leading to modification.

        Todo:
            Add a similar test for Datasets.

        """
        ns_name = 'xsd'
        ns_uri = 'http://www.w3.org/2001/XMLSchema-different'

        with pytest.raises(ValueError) as excinfo:
            iati.utilities.add_namespace(schema_base_tree, ns_name, ns_uri)

        assert 'There is already a namespace called' in str(excinfo.value)

    @pytest.mark.parametrize("not_a_tree", iati.tests.utilities.generate_test_types([], True))
    def test_add_namespace_no_schema(self, not_a_tree):
        """Check that attempting to add a namespace to something that isn't a Schema raises an error."""
        ns_name = 'xsd'
        ns_uri = 'http://www.w3.org/2001/XMLSchema'

        with pytest.raises(TypeError) as excinfo:
            iati.utilities.add_namespace(not_a_tree, ns_name, ns_uri)

        assert 'The `tree` parameter must be of type `etree._ElementTree` - it was of type' in str(excinfo.value)

    @pytest.mark.parametrize("ns_name", iati.tests.utilities.generate_test_types(['str'], True))
    def test_add_namespace_nsname_non_str(self, ns_name):
        """Check that attempting to add a namespace with a name that is not a string acts correctly.

        Todo:
            Determine whether to attempt cast to string or raise an error.

        """
        pass

    @pytest.mark.parametrize("ns_name", [''])
    def test_add_namespace_nsname_invalid_str(self, ns_name, schema_base_tree):
        """Check that attempting to add a namespace with a name that is an invalid string raises an appropriate error.

        Todo:
            Add more tests - for syntax, see:
                https://www.w3.org/TR/REC-xml-names/#NT-NSAttName

        """
        ns_uri = 'http://www.w3.org/2001/XMLSchema'

        with pytest.raises(ValueError) as excinfo:
            iati.utilities.add_namespace(schema_base_tree, ns_name, ns_uri)

        assert 'The `new_ns_name` parameter must be a non-empty string.' in str(excinfo.value)

    @pytest.mark.parametrize("ns_uri", iati.tests.utilities.generate_test_types(['str'], True))
    def test_add_namespace_nsuri_non_str(self, ns_uri):
        """Check that attempting to add a namespace uri that is not a string acts correctly.

        Todo:
            Determine whether to attempt cast to string or raise an error.
        """
        pass

    @pytest.mark.parametrize("ns_uri", [''])
    def test_add_namespace_nsuri_invalid_str(self, ns_uri, schema_base_tree):
        """Check that attempting to add a namespace that is an invalid string raises an appropriate error.

        Note:
            While a valid URI, an empty string is not a valid namespace - https://www.w3.org/TR/REC-xml-names/#iri-use https://www.ietf.org/rfc/rfc2396.txt

        Todo:

        """
        ns_name = 'testname'

        with pytest.raises(ValueError) as excinfo:
            iati.utilities.add_namespace(schema_base_tree, ns_name, ns_uri)

        assert 'The `new_ns_uri` parameter must be a valid URI.' in str(excinfo.value)

    def test_convert_tree_to_schema(self):
        """Check that an etree can be converted to a schema."""
        path = iati.resources.get_schema_path('iati-activities-schema')

        tree = iati.resources.load_as_tree(path)
        if not tree:  # pragma: no cover
            assert False
        schema = iati.utilities.convert_tree_to_schema(tree)

        assert isinstance(schema, etree.XMLSchema)

    def test_convert_xml_to_tree(self):
        """Check that a valid XML string can be converted to an etree."""
        xml = '<parent><child /></parent>'

        tree = iati.utilities.convert_xml_to_tree(xml)

        assert isinstance(tree, etree._Element)  # pylint: disable=protected-access
        assert tree.tag == 'parent'
        assert len(tree.getchildren()) == 1
        assert tree.getchildren()[0].tag == 'child'
        assert not tree.getchildren()[0].getchildren()

    def test_convert_xml_to_tree_invalid_str(self):
        """Check that an invalid string raises an error when an attempt is made to convert it to an etree."""
        not_xml = "this is not XML"

        with pytest.raises(etree.XMLSyntaxError) as excinfo:
            iati.utilities.convert_xml_to_tree(not_xml)

        assert excinfo.typename == 'XMLSyntaxError'

    @pytest.mark.parametrize("not_xml", iati.tests.utilities.generate_test_types(['str'], True))
    def test_convert_xml_to_tree_not_str(self, not_xml):
        """Check that an invalid string raises an error when an attempt is made to convert it to an etree."""
        with pytest.raises(ValueError) as excinfo:
            iati.utilities.convert_xml_to_tree(not_xml)

        assert 'To parse XML into a tree, the XML must be a string, not a' in str(excinfo.value)

    def test_log(self):
        """TODO: Implement testing for logging."""
        pass

    def test_log_error(self):
        """TODO: Implement testing for logging."""
        pass

    def test_log_exception(self):
        """TODO: Implement testing for logging."""
        pass

    def test_log_warning(self):
        """TODO: Implement testing for logging."""
        pass


class TestDefaultVersions(object):
    """A container for tests relating to default versions."""

    @pytest.mark.parametrize("major_version", iati.constants.STANDARD_VERSIONS_MAJOR)
    def test_versions_for_integer(self, major_version):
        """Check that the each of the decimal versions returned by versions_for_integer starts with the input major version."""
        result = iati.utilities.versions_for_integer(major_version)

        for version in result:
            assert version.startswith(str(major_version))



"""A module to provide a copy of all the default data within the IATI SSOT.

This includes Codelists, Schemas and Rulesets at various versions of the Standard.

Todo:
    Handle multiple versions of the Standard rather than limiting to the latest.
    Implement more than Codelists.
"""

import json
import os
from collections import defaultdict
from copy import deepcopy
import iati.codelists
import iati.constants
import iati.resources


def get_default_version_if_none(version):
    """Return the default version number if the input version is None. Otherwise returns the input version as is.

    Args:
        version (str or None): The version to test against.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    Returns:
        str: The default version if the input version is None. Otherwise returns the input version.

    Note:
        The default version of the Standard is deemed to be the latest version.

    """
    if version is None:
        return iati.constants.STANDARD_VERSION_LATEST
    elif version not in iati.constants.STANDARD_VERSIONS:
        raise ValueError("Version {0} is not a valid version of the IATI Standard.".format(version))
    return version


_CODELISTS = defaultdict(dict)
"""A cache of loaded Codelists.

This removes the need to repeatedly load a Codelist from disk each time it is accessed.

The dictionary is structured as:

{
    "version_number_a": {
        "codelist_name_1": iati.Codelist(codelist_1),
        "codelist_name_2": iati.Codelist(codelist_2)
        [...]
    },
    "version_number_b": {
        [...]
    },
    [...]
}

Warning:
    Modifying values directly obtained from this cache can potentially cause unexpected behavior. As such, it is highly recommended to perform a `deepcopy()` on any accessed Codelist before it is modified in any way.

"""


def codelist(name, version=None):
    """Return the default Codelist with the specified name for the specified version of the Standard.

    Args:
        name (str): The name of the Codelist to return.
        version (str): The version of the Standard to return the Codelists for. Defaults to None. This means that the latest version of the specified Codelist is returned.

    Raises:
        ValueError: When a specified name is not a Codelist at the specified version of the Standard.
        ValueError: When a specified version is not a valid version of the Standard.

    Returns:
        iati.Codelist: A Codelist with the specified name from the specified version of the Standard. It is populated with all the Codes on the Codelist.

    Warning:
        A name may not be sufficient to act as a UID.

        Further exploration needs to be undertaken in how to handle multiple versions of the Standard.

    Todo:
        Better distinguish the types of ValueError.

        Better distinguish TypeErrors from KeyErrors - sometimes the latter is raised when the former should have been.

    """
    try:
        codelist_found = _codelists(version, True)[name]
        return deepcopy(codelist_found)
    except (KeyError, TypeError):
        msg = "There is no default Codelist in version {0} of the Standard with the name {1}.".format(version, name)
        iati.utilities.log_warning(msg)
        raise ValueError(msg)


def _codelists(version=None, use_cache=False):
    """Locate the default Codelists for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Codelists for. Defaults to None. This means that the latest version of the Codelists are returned.
        use_cache (bool): Whether the cache should be used rather than loading the Codelists from disk again. If used, a `deepcopy()` should be performed on any returned Codelist before it is modified.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    Returns:
        dict: A dictionary containing all the Codelists at the specified version of the Standard. All Non-Embedded Codelists are included. Keys are Codelist names. Values are iati.Codelist() instances.

    Warning:
        Setting `use_cache` to `True` is dangerous since it does not return a deep copy of the Codelists. This means that modification of a returned Codelist will modify the Codelist everywhere.
        A `deepcopy()` should be performed on any returned value before it is modified.

    Note:
        This is a private function so as to prevent the (dangerous) `use_cache` parameter being part of the public API.

    """
    version = get_default_version_if_none(version)

    paths = iati.resources.get_all_codelist_paths(version)

    for path in paths:
        _, filename = os.path.split(path)
        name = filename[:-len(iati.resources.FILE_CODELIST_EXTENSION)]  # Get the name of the codelist, without the '.xml' file extension
        if (name not in _CODELISTS[version].keys()) or not use_cache:
            xml_str = iati.resources.load_as_string(path)
            codelist_found = iati.Codelist(name, xml=xml_str)
            _CODELISTS[version][name] = codelist_found

    return _CODELISTS[version]


def codelists(version=None):
    """Return the default Codelists for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Codelists for. Defaults to None. This means that the latest version of the Codelists are returned.

    Returns:
        dict: A dictionary containing all the Codelists at the specified version of the Standard. All Non-Embedded Codelists are included. Keys are Codelist names. Values are iati.Codelist() instances, populated with the relevant Codes.

    """
    return _codelists(version)


def codelist_mapping(version=None):
    """Define the mapping process which states where in a Dataset you should find values on a given Codelist.

    Args:
        version (str): The version of the Standard to return the mapping file for. Defaults to None. This means that the mapping file is returned for the latest version of the Standard.

    Returns:
        dict of dict: A dictionary containing mapping information. Keys in the first dictionary are Codelist names. Keys in the second dictionary are `xpath` and `condition`. The condition is `None` if there is no condition.

    Todo:
        Make use of the `version` parameter.

    """
    path = iati.resources.get_codelist_mapping_path(version)
    mapping_tree = iati.resources.load_as_tree(path)
    mappings = defaultdict(list)

    for mapping in mapping_tree.getroot().xpath('//mapping'):
        codelist_name = mapping.find('codelist').attrib['ref']
        codelist_location = mapping.find('path').text

        try:
            condition = mapping.find('condition').text
        except AttributeError:  # there is no condition
            condition = None

        mappings[codelist_name].append({
            'xpath': codelist_location,
            'condition': condition
        })

    return mappings


def ruleset(version=None):
    """Return the Standard Ruleset for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Ruleset for. Defaults to None. This means that the latest Standard Ruleset is returned.

    Returns:
        iati.Ruleset: The default Ruleset for the specified version of the Standard.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    """
    path = iati.resources.get_ruleset_path(iati.resources.FILE_RULESET_STANDARD_NAME, version)
    ruleset_str = iati.resources.load_as_string(path)

    return iati.Ruleset(ruleset_str)


def ruleset_schema(version=None):
    """Return the Ruleset schema for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Ruleset for. Defaults to None. This means that the latest Ruleset schema is returned.

    Returns:
        dict: A dictionary representing the Ruleset schema for the specified version of the Standard.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    """
    path = iati.resources.get_ruleset_path(iati.resources.FILE_RULESET_SCHEMA_NAME, version)
    schema_str = iati.resources.load_as_string(path)

    return json.loads(schema_str)


_SCHEMAS = defaultdict(lambda: defaultdict(dict))
"""A cache of loaded Schemas.

This removes the need to repeatedly load a Schema from disk each time it is accessed.

{
    "version_number_a": {
        "populated": {
            "iati-activities": iati.ActivitySchema
            "iati-organisations": iati.OrganisationSchema
        },
        "unpopulated": {
            "iati-activities": iati.ActivitySchema
            "iati-organisations": iati.OrganisationSchema
        },
    },
    "version_number_b": {
        [...]
    },
    [...]
}

Warning:
    Modifying values directly obtained from this cache can potentially cause unexpected behavior. As such, it is highly recommended to perform a `deepcopy()` on any accessed Schema before it is modified in any way.

"""


def _populate_schema(schema, version=None):
    """Populate a Schema with all its extras.

    The extras include Codelists and Rulesets.

    Args:
        schema (iati.Schema): The Schema to populate.
        version (str): The version of the Standard to return the Schema for. Defaults to None. This means that the latest version of the Standard is assumed.

    Returns:
        iati.Schema: The provided Schema, populated with additional information.

    Warning:
        Does not create a copy of the provided Schema, instead adding to it directly.

    """
    version = get_default_version_if_none(version)

    codelists_to_add = codelists(version)
    for codelist_to_add in codelists_to_add.values():
        schema.codelists.add(codelist_to_add)

    schema.rulesets.add(ruleset(version))

    return schema


def _schema(path_func, schema_class, version=None, populate=True, use_cache=False):
    """Return the default Schema of the specified type for the specified version of the Standard.

    Args:
        path_func (func): A function to return the paths at which the relevant Schema can be found.
        schema_class (type): A class definition for the Schema of interest.
        version (str): The version of the Standard to return the Schema for. Defaults to None. This means that the latest version of the Schema is returned.
        populate (bool): Whether the Schema should be populated with auxilliary information such as Codelists and Rulesets.
        use_cache (bool): Whether the cache should be used rather than loading the Schema from disk again. If used, a `deepcopy()` should be performed on any returned Schema before it is modified.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    Returns:
        iati.Schema: An instantiated IATI Schema for the specified version.

    """
    population_key = 'populated' if populate else 'unpopulated'

    version = get_default_version_if_none(version)

    schema_paths = path_func(version)

    if (schema_class.ROOT_ELEMENT_NAME not in _SCHEMAS[version][population_key].keys()) or not use_cache:
        schema = schema_class(schema_paths[0])
        if populate:
            schema = _populate_schema(schema, version)
        _SCHEMAS[version][population_key][schema_class.ROOT_ELEMENT_NAME] = schema

    return _SCHEMAS[version][population_key][schema_class.ROOT_ELEMENT_NAME]


def activity_schema(version=None, populate=True):
    """Return the default Activity Schema for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Schema for. Defaults to None. This means that the latest version of the Schema is returned.
        populate (bool): Whether the Schema should be populated with auxilliary information such as Codelists and Rulesets.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    Returns:
        iati.ActivitySchema: An instantiated IATI Schema for the specified version of the Standard.

    """
    return _schema(iati.resources.get_all_activity_schema_paths, iati.ActivitySchema, version, populate)


def organisation_schema(version=None, populate=True):
    """Return the default Organisation Schema for the specified version of the Standard.

    Args:
        version (str): The version of the Standard to return the Schema for. Defaults to None. This means that the latest version of the Schema is returned.
        populate (bool): Whether the Schema should be populated with auxilliary information such as Codelists and Rulesets.

    Raises:
        ValueError: When a specified version is not a valid version of the IATI Standard.

    Returns:
        iati.OrganisationSchema: An instantiated IATI Schema for the specified version of the Standard.

    """
    return _schema(iati.resources.get_all_organisation_schema_paths, iati.OrganisationSchema, version, populate)
